from typing import Iterator, Tuple, Optional, Union

import torch
from torch import nn, Tensor, device
import torch._dynamo as dynamo
from torch.nn.parallel import DistributedDataParallel as DDP

from ..core import Processor
from ..utils import recursive_clone, to_device, recursive_device
from ..path import beam_path, local_copy


class BeamNN(nn.Module, Processor):
    """
    BeamNN is a wrapper class around PyTorch's nn.Module, with added functionalities
    from the Processor class. It allows dynamic integration of an existing nn.Module
    instance with additional processing capabilities.

    Attributes:
        _sample_input: A sample input for JIT tracing or other optimization methods.
        _module: The wrapped nn.Module instance.
    """

    def __init__(self, *args, _module=None, _model_type=None, **kwargs):

        """
        Initialize the BeamNN wrapper.

        Args:
            _module (nn.Module): The PyTorch module to be wrapped.
            *args, **kwargs: Additional arguments for the Processor initialization.
        """

        nn.Module.__init__(self)
        Processor.__init__(self, *args, **kwargs)
        self._sample_input = None
        self._sample_output = None
        self._module = _module
        self._model_type = _model_type or 'torch'

    @classmethod
    def from_module(cls, module, *args, hparams=None, **kwargs):

        """
        Class method to create a BeamNN object from an existing nn.Module.

        Args:
            module (nn.Module): The PyTorch module to be wrapped.

        Returns:
            BeamNN: A new BeamNN instance wrapping the provided module.
        """

        if isinstance(module, cls):
            if hparams is not None:
                module.update_hparams(hparams)
            beam_module = module
        else:
            beam_module = cls(*args, _module=module, hparams=hparams, **kwargs)
        return beam_module

    def __dir__(self):
        d = super().__dir__()
        if self.module_exists:
            d.extend(self._module.__dir__())
        return d

    def _mixin_method(self, method_name, *args, **kwargs):
        if self.module_exists:
            return getattr(self._module, method_name)(*args, **kwargs)
        return getattr(nn.Module, method_name)(self, *args, **kwargs)

    def parameters(self, *args, **kwargs):
        return self._mixin_method('parameters', *args, **kwargs)

    def named_parameters(self, *args, **kwargs):
        return self._mixin_method('named_parameters', *args, **kwargs)

    def state_dict(self, *args, **kwargs):
        return self._mixin_method('state_dict', *args, **kwargs)

    def load_state_dict(self, state_dict, *args, **kwargs):
        return self._mixin_method('load_state_dict', state_dict, *args, **kwargs)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        return self._mixin_method('_load_from_state_dict', state_dict, prefix, local_metadata, strict,
                                  missing_keys, unexpected_keys, error_msgs)

    def apply(self, fn):
        return self._mixin_method('apply', fn)

    def register_buffer(self, name, tensor, persistent=True):
        return self._mixin_method('register_buffer', name, tensor, persistent)

    def register_parameter(self, name, param):
        return self._mixin_method('register_parameter', name, param)

    def extra_repr(self):
        return self._mixin_method('extra_repr')

    def children(self):
        return self._mixin_method('children')

    def modules(self):
        return self._mixin_method('modules')

    def _modules(self):
        return self._mixin_method('_modules')

    def _load_state_dict_post_hooks(self):
        hook_dict = self._mixin_method('_load_state_dict_post_hooks')
        if self.module_exists:

            for k, fn in hook_dict.items():
                def wrapper(module, *args, **kwargs):
                    return fn(module._module, *args, **kwargs)

                hook_dict[k] = wrapper

        return hook_dict

    def named_modules(self, *args, **kwargs):
        return self._mixin_method('named_modules', *args, **kwargs)

    def named_children(self) -> Iterator[Tuple[str, nn.Module]]:
        return self._mixin_method('named_children')

    def get_parameter(self, target: str) -> nn.Parameter:
        return self._mixin_method('get_parameter', target)

    def get_buffer(self, target: str) -> torch.Tensor:
        return self._mixin_method('get_buffer', target)

    def get_submodule(self, target: str) -> nn.Module:
        return self._mixin_method('get_submodule', target)

    def buffers(self, recurse: bool = True) -> Iterator[Tensor]:
        return self._mixin_method('buffers', recurse)

    def named_buffers(self, *arg, **kwargs) -> Iterator[Tuple[str, Tensor]]:
        return self._mixin_method('named_buffers', *arg, **kwargs)

    def register_module(self, *args, **kwargs) -> None:
        return self._mixin_method('register_module', *args, **kwargs)

    def to_empty(self, *args, **kwargs):
        return self._mixin_method('to_empty', *args, **kwargs)

    def add_module(self, name: str, module: Optional[nn.Module]) -> None:
        return self._mixin_method('add_module', name, module)

    def to(self, *args, **kwargs):
        if self.module_exists:
            self._module = self._module.to(*args, **kwargs)
        else:
            nn.Module.to(self, *args, **kwargs)
        return self

    def cuda(self, device: Optional[Union[int, device]] = None) -> nn.Module:
        if self.module_exists:
            self._module = self._module.cuda(device)
        else:
            nn.Module.cuda(self, device)
        return self

    def cpu(self) -> nn.Module:
        if self.module_exists:
            self._module = self._module.cpu()
        else:
            nn.Module.cpu(self)
        return self

    def type(self, dst_type: Optional[Union[str, device]] = None, *args, **kwargs) -> nn.Module:
        if self.module_exists:
            self._module = self._module.type(dst_type, *args, **kwargs)
        else:
            nn.Module.type(self, dst_type, *args, **kwargs)
        return self

    def float(self) -> nn.Module:
        if self.module_exists:
            self._module = self._module.float()
        else:
            nn.Module.float(self)
        return self

    def double(self) -> nn.Module:
        if self.module_exists:
            self._module = self._module.double()
        else:
            nn.Module.double(self)
        return self

    def half(self) -> nn.Module:
        if self.module_exists:
            self._module = self._module.half()
        else:
            nn.Module.half(self)
        return self

    def bfloat16(self) -> nn.Module:
        if self.module_exists:
            self._module = self._module.bfloat16()
        else:
            nn.Module.bfloat16(self)
        return self

    def __repr__(self):
        if self.module_exists:
            module_repr = repr(self._module)
            return f"BeamNN(wrapper)(\n{module_repr})"
        return f"BeamNN:{nn.Module.__repr__(self)}"

    @property
    def module_exists(self):
        if not hasattr(self, '_module') or self._module is None:
            return False
        return True

    def forward(self, *args, **kwargs):
        if self.module_exists:
            return self._module.forward(*args, **kwargs)
        raise NotImplementedError("Implement forward method in your BeamNN subclass")

    def __getattr__(self, item):
        if item != '_module' and self.module_exists:
            return getattr(self._module, item)
        return nn.Module.__getattr__(self, item)

    def __setattr__(self, key, value):
        if self.module_exists and not hasattr(self, key):
            return setattr(self._module, key, value)
        return nn.Module.__setattr__(self, key, value)

    @dynamo.disable
    def save_sample_input(self, *args, **kwargs):

        self._sample_input = {'args': recursive_clone(to_device(args, device='cpu')) if args else None,
                              'kwargs': recursive_clone(to_device(kwargs, device='cpu')) if kwargs else None,
                              'args_device': recursive_device(args) if args else None,
                              'kwargs_device': recursive_device(kwargs) if kwargs else None}

    @dynamo.disable
    def save_sample_output(self, output):
        self._sample_output = {'output': recursive_clone(to_device(output, device='cpu')),
                               'output_device': recursive_device(output)}

    def __call__(self, *args, **kwargs):

        if self._sample_input is None:
            self.save_sample_input(*args, **kwargs)

        if self.module_exists:
            output = self._module(*args, **kwargs)
        else:
            output = nn.Module.__call__(self, *args, **kwargs)

        if self._sample_output is None:
            self.save_sample_output(output)

        return output

    def optimize(self, method='compile', **kwargs):
        if method == 'compile':
            return BeamNN._compile(self, **kwargs)
        elif method == 'jit_trace':
            return BeamNN._jit_trace(self, **kwargs)
        elif method == 'jit_script':
            return BeamNN._jit_script(self)
        else:
            raise ValueError(f'Invalid optimization method: {method}, must be one of [compile|jit_trace|jit_script]')

    def export(self, path, method='onnx', **kwargs):
        if method == 'onnx':
            self._export_onnx(path, **kwargs)
        elif method == 'torchscript':
            self._export_jit_script(self)
        elif method == 'tensorrt':
            self._export_trt(path, **kwargs)
        else:
            raise ValueError(f'Invalid export method: {method}, must be one of [onnx]')

    def export_triton_server(self, path, method='onnx', model_name=None, model_version=None, **kwargs):

        from .triton import TritonConfig

        if model_version is None:
            model_version = '1'
        path = beam_path(path)
        if model_name is not None:
            path = path.joinpath(model_name)
        model_name = path.name
        path = path.joinpath(model_version)
        path.mkdir(parents=True, exist_ok=True)
        self.export(path.joinpath('model.onnx'), method=method, **kwargs)

        example_inputs, example_kwarg_inputs = self.example_input
        if example_inputs:
            max_batch_size = example_inputs[0].shape[0]
        elif example_kwarg_inputs:
            max_batch_size = next(example_kwarg_inputs.values()).shape[0]
        else:
            max_batch_size = 1

        config = TritonConfig(name=model_name, platform='pytorch_libtorch', max_batch_size=max_batch_size)
        config.save_to_file(path.joinpath('config.pbtxt'))

    @property
    def sample_input(self):
        return self._sample_input

    @property
    def sample_output(self):
        return self._sample_output

    @property
    def example_input(self):

        if self.sample_input['args'] is not None:
            example_inputs = recursive_clone(to_device(self.sample_input['args'],
                                                       device=self.sample_input['args_device']))
        else:
            example_inputs = None
        if self.sample_input['kwargs'] is not None:
            example_kwarg_inputs = recursive_clone(to_device(self.sample_input['kwargs'],
                                                         device=self.sample_input['kwargs_device']))
        else:
            example_kwarg_inputs = None
        return example_inputs, example_kwarg_inputs

    @classmethod
    def _jit_trace(cls, self, check_trace=None, check_inputs=None, check_tolerance=None, strict=None):

        self.assert_forward_was_executed('jit_trace')
        check_trace = check_trace or self.get_hparam('jit_check_trace', True)
        check_inputs = check_inputs or self.get_hparam('jit_check_inputs', None)
        check_tolerance = check_tolerance or self.get_hparam('jit_check_tolerance', 1e-5)
        strict = strict or self.get_hparam('jit_strict', True)

        module = self._module or self

        example_inputs, example_kwarg_inputs = self.example_input
        optimized = torch.jit.trace(module, example_inputs=example_inputs,
                               check_trace=check_trace, check_inputs=check_inputs, check_tolerance=check_tolerance,
                               strict=strict, example_kwarg_inputs=example_kwarg_inputs)

        return cls(_module=optimized, hparams=self.hparams, _model_type='torchscript')

    @classmethod
    def _jit_script(cls, self):

        self.assert_forward_was_executed('jit_script')
        example_inputs, example_kwarg_inputs = self.example_input
        if example_kwarg_inputs:
            raise NotImplementedError("JIT script does not support keyword arguments")

        module = self._module or self
        optimized = torch.jit.script(module, example_inputs=[example_inputs])

        return cls(_module=optimized, hparams=self.hparams, _model_type='torchscript')

    def assert_forward_was_executed(self, method_name):
        if self.sample_output is None or self.sample_input is None:
            raise ValueError(f"Please call the forward method at least once before calling {method_name}")

    @classmethod
    def _compile(cls, self, fullgraph=None, dynamic=None, backend=None,
                 mode=None, options=None, disable=False):

        fullgraph = fullgraph or self.get_hparam('compile_fullgraph', None)
        dynamic = dynamic or self.get_hparam('compile_dynamic', False)
        backend = backend or self.get_hparam('compile_backend', 'inductor')
        mode = mode or self.get_hparam('compile_mode', None)
        options = options or self.get_hparam('compile_options', None)

        module = self._module or self
        optimized = torch.compile(module, fullgraph=fullgraph, dynamic=dynamic, backend=backend,
                             mode=mode, options=options, disable=disable)

        return cls(_module=optimized, hparams=self.hparams, _model_type='compiled')

    def _export_jit_script(self, path):

        path = beam_path(path)
        with local_copy(path, as_beam_path=False) as tmp_path:
            if self._model_type == 'torchscript':
                self.module.save_torchscript(tmp_path)
            elif self._model_type == 'torch':
                model = self._jit_script(self).save(tmp_path)
                model.save_torchscript(tmp_path)
            else:
                raise NotImplementedError("Only TorchScript or native pytorch models can be exported with this method")

    def save_torchscript(self):
        if self.module_exists and self._model_type == 'torchscript':
            return self._module.save()
        raise NotImplementedError("Only TorchScript models can be saved with this method")

    def _export_trt(self, path,**kwargs):

        self._export_onnx(path, **kwargs)

        import tensorrt as trt
        TRT_LOGGER = trt.Logger(trt.Logger.WARNING)

        # Create a builder
        builder = trt.Builder(TRT_LOGGER)
        # Specify the builder configurations
        network = builder.create_network()
        parser = trt.OnnxParser(network, TRT_LOGGER)

        with local_copy(path, as_beam_path=True) as tmp_path:
            # Parse the model to create a network.
            parser.parse(tmp_path.read())
            # Build the engine
            engine = builder.build_cuda_engine(network)
            tmp_path.write_bin(engine.serialize())

    def _export_onnx(self, path, export_params=True, verbose=False, training='eval',
                     input_names=None, output_names=None, operator_export_type='ONNX', opset_version=None,
                     do_constant_folding=True, dynamic_axes=None, keep_initializers_as_inputs=None, custom_opsets=None,
                     export_modules_as_functions=False):

        self.assert_forward_was_executed('export(\'onnx\')')
        example_inputs, _ = self.example_input

        import torch.onnx

        if training == 'eval':
            training = torch.onnx.TrainingMode.EVAL
        elif training == 'train':
            training = torch.onnx.TrainingMode.TRAINING
        else:
            training = torch.onnx.TrainingMode.PRESERVE

        if operator_export_type == 'ONNX':
            operator_export_type = torch.onnx.OperatorExportTypes.ONNX
        elif operator_export_type == 'ONNX_ATEN':
            operator_export_type = torch.onnx.OperatorExportTypes.ONNX_ATEN
        elif operator_export_type == 'ONNX_FALLTHROUGH':
            operator_export_type = torch.onnx.OperatorExportTypes.ONNX_FALLTHROUGH
        else:
            raise ValueError(f'Invalid operator_export_type: {operator_export_type}, '
                             f'must be one of "ONNX", "ONNX_ATEN", or "ONNX_FALLTHROUGH"')

        path = beam_path(path)
        module = self._module or self
        with local_copy(path, as_beam_path=False) as tmp_path:
            torch.onnx.export(module, example_inputs, tmp_path, export_params=export_params,
                                     verbose=verbose, training=training, input_names=input_names,
                                     output_names=output_names, operator_export_type=operator_export_type,
                                     opset_version=opset_version, do_constant_folding=do_constant_folding,
                                     dynamic_axes=dynamic_axes, keep_initializers_as_inputs=keep_initializers_as_inputs,
                                     custom_opsets=custom_opsets,
                                     export_modules_as_functions=export_modules_as_functions)

    # add pruning and quantization methods
    # add methods for converting to other frameworks?


class BeamDDP(DDP):

    def __init__(self, module, *args, **kwargs):
        super().__init__(module, *args, **kwargs)
        self._init_is_done = True

    def __getattr__(self, item):
        if item.startswith('_') or item == '_init_is_done' or not hasattr(self, '_init_is_done'):
            return super().__getattr__(item)
        return getattr(self.module, item)




from ..config import boolean_feature, basic_beam_parser


def get_ssl_parser():

    parser = basic_beam_parser()

    boolean_feature(parser, "verbose-lgb", False, "Print progress in lgb training")
    parser.add_argument('--similarity', type=str, metavar='hparam', default='cosine',
                        help='Similarity distance in UniversalSSL')
    parser.add_argument('--p-dim', type=int, default=None, help='Prediction/Projection output dimension')
    parser.add_argument('--temperature', type=float, default=1.0, metavar='hparam', help='Softmax temperature')
    parser.add_argument('--var-eps', type=float, default=0.0001, metavar='hparam', help='Std epsilon in VICReg')
    parser.add_argument('--lambda-vicreg', type=float, default=25., metavar='hparam',
                        help='Lambda weight in VICReg')
    parser.add_argument('--mu-vicreg', type=float, default=25., metavar='hparam', help='Mu weight in VICReg')
    parser.add_argument('--nu-vicreg', type=float, default=1., metavar='hparam', help='Nu weight in VICReg')
    parser.add_argument('--lambda-mean-vicreg', type=float, default=20., metavar='hparam',
                        help='lambda-mean weight in BeamVICReg')
    parser.add_argument('--tau', type=float, default=.99, metavar='hparam', help='Target update factor')
    parser.add_argument('--lambda-twins', type=float, default=0.005, metavar='hparam',
                        help='Off diagonal weight factor for Barlow Twins loss')

    parser.add_argument('--lgb-rounds', type=int, default=40, help='LGB argument: num_round')
    parser.add_argument('--lgb-num-leaves', type=int, default=31, help='LGB argument: num_leaves')
    parser.add_argument('--lgb-max-depth', type=int, default=4, help='LGB argument: max_depth')
    parser.add_argument('--lgb-device', type=int, default=None, help='LGB argument: device')
    return parser



import json

from ..llm import beam_llm
from ..core import MetaDispatcher
from ..utils import lazy_property, check_type
from ..logger import beam_logger as logger
import inspect
import threading
import queue
import re

# do not remove the following imports
import torch
import numpy as np
import pandas as pd
# do not remove the above imports


class BeamAssistant(MetaDispatcher):

    def __init__(self, obj, *args, llm=None, llm_kwargs=None, summary_len=10, eval_arguments=True, retries=3,
                 _summary=None, **kwargs):

        super().__init__(obj, *args, summary_len=summary_len,
                         eval_arguments=eval_arguments, retries=retries, **kwargs)

        self.summary_len = self.get_hparam('summary_len', summary_len)
        self.eval_arguments = self.get_hparam('eval_arguments', eval_arguments)
        self.retries = self.get_hparam('retries', retries)
        llm = self.get_hparam('llm', llm)
        llm_kwargs = llm_kwargs or {}

        self.llm = None
        if llm is not None:
            self.llm = beam_llm(llm, **llm_kwargs)

        self._summary_queue = queue.Queue()
        self._summarize_thread = None
        if _summary is not None:
            self._summary_queue.put(_summary)
        else:
            # summarize the object in a different thread
            self._summarize_thread = threading.Thread(target=self.summarize, daemon=True)
            self._summarize_thread.start()

    @lazy_property
    def summary(self):
        if self._summarize_thread is not None:
            # wait for the thread to finish and get the summary from the queue
            self._summarize_thread.join()
        return self._summary_queue.get()

    @lazy_property
    def doc(self):
        return self.real_object.__doc__

    @lazy_property
    def source(self):
        if self.type in ['class', 'instance']:
            # iterate over all parent classes and get the source
            sources = []
            base_cls = self.real_object if self.type == 'class' else self.real_object.__class__
            for cls in inspect.getmro(base_cls):
                if cls.__module__ != 'builtins':
                    sources.append(inspect.getsource(cls))
            # sum all the sources
            return '\n'.join(sources)
        else:
            return inspect.getsource(self.real_object)

    @property
    def name(self):
        if self.type == 'class':
            return self.real_object.__name__
        elif self.type == 'instance':
            return self.real_object.__class__.__name__
        else:
            return self.real_object.__name__

    @lazy_property
    def type_name(self):
        if self.type == 'class':
            return 'class'
        elif self.type == 'instance':
            return 'class instance'
        elif self.type == 'function':
            return 'function'
        elif self.type == 'method':
            return 'class method'
        else:
            raise ValueError(f"Unknown type: {self.type}")

    def summarize(self, **kwargs):
        prompt = (f"Summarize the {self.type_name}: {self.name} with up to {self.summary_len} sentences "
                  f"given the following source code:\n\n{self.source}\n"
                  f"Your answer:\n\n")

        summary = self.ask(prompt, system=False, **kwargs).text
        # put the summary in the queue
        self._summary_queue.put(summary)

    def ask(self, query, system=True, **kwargs):
        if system:
            query = f"{self.system_prompt}\n{query}"
        else:
            query = f"{query}"
        return self.llm.ask(query, **kwargs)

    @lazy_property
    def system_prompt(self):
        return (f"Your job is to help a programmer to execute a python code from natural language queries.\n"
                f"You are given a {self.type_name} named {self.name} with the following description:\n"
                f"{self.summary}\n")

    def init_instance(self, query, ask_kwargs=None, user_kwargs=None, eval_arguments=None):

        eval_arguments = eval_arguments or self.eval_arguments
        user_kwargs = user_kwargs or {}

        # create an instance of the class and return a NLPDispatcher object
        def is_class_method(member):
            # First, ensure that member is a method bound to a class
            if inspect.ismethod(member) and inspect.isclass(member.__self__):
                # Now that we've confirmed member is a method, check the name conditions
                if not member.__name__.startswith('_') and member.__name__ != 'from_nlp':
                    return True
            return False

        classmethods = [name for name, member in inspect.getmembers(self.real_object, predicate=is_class_method)]

        example_output = {'method': 'method_name'}
        query = (f"Choose the best classmethod that should be used to build a class instance according to the "
                  f"following query:\n"
                  f"Query: {query}\n"
                  f"Methods: {classmethods}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json.dumps(example_output)}\n"
                  f"Your answer:\n\n")

        ask_kwargs = ask_kwargs or {}
        response = self.ask(query, **ask_kwargs).json

        constructor_name = response['method']

        if constructor_name not in classmethods:
            raise ValueError(f"Constructor {constructor_name} not found in the list of class constructors")

        constructor = getattr(self.real_object, constructor_name)
        logger.info(f"Using classmethod {constructor_name} to build the class instance")

        constructor_sourcecode = inspect.getsource(constructor)
        init_sourcecode = inspect.getsource(self.real_object.__init__)

        json_output_example = {"args": ['arg1', 'arg2'], "kwargs": {'kwarg1': 'value1', 'kwarg2': 'value2'}}
        query = (f"Build a suitable dictionary of arguments and keyword arguments to build a class instance according "
                  f"to the following query:\n"
                  f"Query: {query}\n"
                  f"with the classmethod: {constructor_name} (of class {self.name}) with source-code:\n"
                  f"{constructor_sourcecode}\n"
                  f"and the class __init__ method source-code:\n"
                  f"{init_sourcecode}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json_output_example}\n"
                  f"Your answer:\n\n")

        d = self.ask(query, **ask_kwargs).json
        args = d.get('args', [])
        kwargs = d.get('kwargs', {})

        logger.info(f"Using args: {args} and kwargs: {kwargs} to build the class instance")

        instance = constructor(*args, **kwargs)

        return BeamAssistant(instance, llm=self.llm, summary_len=self.summary_len, _summary=self.summary)

    def exec_method(self, query, method_name=None, ask_kwargs=None, user_kwargs=None, eval_arguments=None):

        eval_arguments = eval_arguments or self.eval_arguments

        assert method_name is not None or self.type == 'function', \
            'method_name must be provided for non-function objects'

        if method_name is None:
            method_description = f"function: {self.name}"
        else:
            method_description = f"class method: {method_name} (of class {self.name})"

        ask_kwargs = ask_kwargs or {}
        user_kwargs = user_kwargs or {}
        args_hint = ''
        if user_kwargs:
            args_hint = (f"These are arguments that the user want to pass to the method "
                         f"(their names represent their meaning but they do not necessarily "
                         f"match to the method signature):\n"
                         f"{self.arguments_repr(**user_kwargs)}\n")

        if method_name is not None:
            method = getattr(self.real_object, method_name)
        else:
            method = self.real_object

        sourcecode = inspect.getsource(method)

        json_output_example = {"args": [2, 'some_str'],
                               "kwargs": {'k1': 3.5, 'k2': None}}

        if eval_arguments:
            json_output_example["args"].append('<eval>np.arange(100)</eval>')
            json_output_example["kwargs"]['ke'] = "<eval>torch.randn(20)</eval>"

        if args_hint:
            json_output_example["args"].append('<user_arg>arg_name</user_arg>')
            json_output_example["kwargs"]['ku'] = "<user_kwarg>arg_name</user_kwarg>"

        i = 1
        possible_argument_types = ("As argument, you can use the following types:\n"
                                   "1. JSON serializable objects\n"
                                   )

        if eval_arguments:
            i += 1
            possible_argument_types += (f"{i}. python statements (wrapped with <eval></eval> tags). "
                                        "In your statement you can use the following python packages: "
                                        "(1) numpy as np (2) torch (3) pandas as pd\n")

        if args_hint:
            i += 1
            possible_argument_types += f"{i}. User-defined arguments (wrapped with <user_arg></user_arg> tags)\n"

        prompt = (f"Build a suitable dictionary of arguments and keyword arguments to answer the following query:\n"
                  f"Query: {query}\n"
                  f"with the {method_description} with source-code:\n"
                  f"{sourcecode}\n"
                  f"{args_hint}"
                  f"{possible_argument_types}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json.dumps(json_output_example)}\n"
                  f"Your answer:\n\n")

        d = self.ask(prompt, **ask_kwargs).json

        args = d.get('args', [])
        kwargs = d.get('kwargs', {})

        # iterate over args and kwargs and look for tags <eval> and <user_arg>
        # use re to match the pattern <tag>content</tag>

        eval_pattern = re.compile(r"<eval>(.*)</eval>")
        user_arg_pattern = re.compile(r"<user_arg>(.*)</user_arg>")

        def process_value(v):

            if not isinstance(v, str):
                return v

            eval_match = eval_pattern.match(v)
            user_arg_match = user_arg_pattern.match(v)
            if eval_match:
                # extract the content and evaluate it
                try:
                    v = eval(eval_match.group(1))
                except:
                    raise ValueError(f"Failed to evaluate the expression: {eval_match.group(1)}")
            elif user_arg_match:
                # extract the content and use it as a user-defined argument
                v = user_arg_match.group(1)
                v = user_kwargs[v]['value']

            return v

        logger.info(f"Using args: {args} and kwargs: {kwargs} to answer the query")

        for k, v in kwargs.items():
            kwargs[k] = process_value(v)

        for i, v in enumerate(args):
            args[i] = process_value(v)

        return method(*args, **kwargs)

    @staticmethod
    def arguments_repr(**kwargs):
        args_repr = ''
        for k, v in kwargs.items():
            args_repr += f"{k}: {v['str']} (type: {type(v['value'])}, {v['metadata']})\n"
        return args_repr

    def choose_method(self, query, ask_kwargs=None, user_kwargs=None):

        ask_kwargs = ask_kwargs or {}
        user_kwargs = user_kwargs or {}
        args_hint = ''
        if user_kwargs:
            args_hint = (f"These are arguments that the user want to pass to the method "
                         f"(their names do not necessarily match to the method signature):\n"
                         f"{self.arguments_repr(**user_kwargs)}\n")

        method_list = inspect.getmembers(self.real_object, predicate=inspect.isroutine)
        method_list = [m for m in method_list if not m[0].startswith('_')]
        json_output_example = json.dumps({'method': 'method_name'})
        class_doc = inspect.getdoc(self)
        class_doc = f"{class_doc}\n" if class_doc else ""

        prompt = (f"Choose the suitable method that should be used to answer the following query:\n"
                  f"Query: {query}\n"
                  f"Class: {self.__class__.__name__}\n"
                  f"{class_doc}"
                  f"Attributes: {method_list}\n"
                  f"{args_hint}"
                  f"Return your answer as a JSON object, see example below:\n"
                  f"{json_output_example}\n"
                  f"Your answer:\n\n")

        response = self.ask(prompt, **ask_kwargs).json
        method_name = response['method']

        if method_name not in [m[0] for m in method_list]:
            raise ValueError(f"Method {method_name} not found in the list of methods")

        logger.info(f"Using method {method_name} to answer the query")

        return method_name

    def exec(self, query, method=None, ask_kwargs=None, eval_arguments=None, **kwargs):
        # execute code according to the prompt

        user_kwargs = {k: {'value': v, 'str': str(v), 'metadata': check_type(v)} for k, v in kwargs.items()}

        if self.type == 'class':
            # create an instance of the class
            res = self.init_instance(query, ask_kwargs=ask_kwargs, user_kwargs=user_kwargs, eval_arguments=eval_arguments)
        elif self.type == 'instance':
            if method is None:
                method = self.choose_method(query, ask_kwargs=ask_kwargs, user_kwargs=user_kwargs)
            res = self.exec_method(query, method_name=method, ask_kwargs=ask_kwargs,
                                   user_kwargs=user_kwargs, eval_arguments=eval_arguments)
        elif self.type == 'function':
            res = self.exec_method(query, ask_kwargs=ask_kwargs,
                                   user_kwargs=user_kwargs, eval_arguments=eval_arguments)
        else:
            raise ValueError(f"Unknown type: {self.type}")
        return res

    def chat(self, query):
        suggestion = self.ask(query).text
        return suggestion

    def gen_code(self, query):
        code = self.ask(
            f"Return an executable python code that performs the following task:\n"
            f"{query}\\n"
            f"Use a markdown syntax, i.e. the tags \'\'\'python and \'\'\' at the beginning and the end of the code "
            f"section. "
            f"The final result should be assigned to a variable name \'result\'.\n"
            f"Your answer:\n\n").text
        # use re to extract the code
        code = re.search(r"\'\'\'python(.*)\'\'\'", code, re.DOTALL).group(1)
        return code

    def exec_eval(self, query):
        code = self.gen_code(query)
        try:
            exec(code, globals(), locals())
            return result
        except:
            raise ValueError(f"Failed to execute the code:\n{code}")



import mlflow.pyfunc
import mlflow
from mlflow import log_params, log_metric
import os
from ..path import beam_path
from ..algorithm import NeuralAlgorithm


class MLflowSummaryWriter:
    def __init__(self,  exp_name, tensorboard_hparams=None, mlflow_uri=None):

        if mlflow_uri is None:
            mlflow_uri = os.environ['MLFLOW_TRACKING_URI']

        mlflow.set_tracking_uri(mlflow_uri)
        mlflow.set_experiment(exp_name)

        self.mlflow_run = mlflow.start_run(run_name=exp_name)

        if tensorboard_hparams is not None:
            for param_name, param_value in tensorboard_hparams.items():
                mlflow.log_param(param_name, param_value)

        self._url = None

    @property
    def url(self):
        if self._url is None:

            tracking_uri = mlflow.get_tracking_uri()
            current_experiment_id = mlflow.active_run().info.experiment_id
            # Construct the experiment URL (this assumes the default MLflow UI URL structure)
            self._url = f"{tracking_uri}/#/experiments/{current_experiment_id}"

        return self._url

    def add_hparams(self, hparam_dict, metric_dict, **kwargs):
        log_params(hparam_dict)
        for key, value in metric_dict.items():
            log_metric(key, value)

    def add_scalar(self, tag, scalar_value, global_step=None):
        log_metric(tag, scalar_value, step=global_step)

    def add_scalars(self, main_tag, tag_scalar_dict, global_step=None):
        for tag, scalar_value in tag_scalar_dict.items():
            full_tag = f"{main_tag}/{tag}"
            log_metric(full_tag, scalar_value, step=global_step)

    def close(self):
        mlflow.end_run()


class MFBeamAlgWrapper(mlflow.pyfunc.PythonModel):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alg = None

    def load_context(self, context):

        path_to_hparams = context.artifacts['hparams']
        state = context.artifacts['state']

        hparams = beam_path(path_to_hparams).read()

        self.alg = NeuralAlgorithm(hparams)
        self.alg.load_checkpoint(state)

    def predict(self, context, model_input):
        return self.alg.predict(model_input)

    #TODO: follow https://medium.com/@pennyqxr/how-save-and-load-fasttext-model-in-mlflow-format-37e4d6017bf0
    @staticmethod
    def save_model(alg, name, stage=None):

        checkpoint_file = alg.experiment.checkpoints_dir.joinpath(f'checkpoint_mlflow_{alg.epoch + 1:06d}')
        alg.save_checkpoint(checkpoint_file)

        artifacts = {'hparams': str(alg.experiment.experiment_dir.joinpath('hparams.pkl')),
                     'state': str(checkpoint_file)}
        with mlflow.start_run() as run:
            mlflow.pyfunc.log_model(
                artifact_path=str(checkpoint_file.parent.joinpath(checkpoint_file.stem)),
                python_model=alg,
                code_path=[str(alg.experiment.source_dir)],
                artifacts=artifacts,
                registered_model_name=f"{name}/{stage}",

            )

    @staticmethod
    def load_model(name, stage=None):

        if stage is None:
            stage = mlflow.tracking.MlflowClient().get_latest_versions(name, stages=['Production'])[0].version

        loaded_model = mlflow.pyfunc.load_model(model_uri=f"models:/{name}/{stage}")
        return loaded_model


import argparse
import copy
import os
import sys
from argparse import Namespace
import re
from collections import defaultdict

from ..utils import is_notebook, check_type
from ..path import beam_path, beam_key
from ..logger import beam_logger as logger
from .._version import __version__
import re


def boolean_feature(parser, feature, default=False, help='', metavar=None):
    featurename = feature.replace("-", "_")
    feature_parser = parser.add_mutually_exclusive_group(required=False)
    feature_parser.add_argument('--%s' % feature, dest=featurename, action='store_true', help=help)
    feature_parser.add_argument('--no-%s' % feature, dest=featurename, action='store_false', help=help)
    pa = parser._actions
    for a in pa:
        if a.dest == featurename:
            a.metavar = metavar
    parser.set_defaults(**{featurename: default})


def empty_beam_parser():

    parser = argparse.ArgumentParser(description='List of available arguments for this project',
                                     conflict_handler='resolve')
    parser.add_argument('config_files', nargs='*',
                        help='A list config files (optional) to load, will be loaded in reverse order '
                             '(first one has higher priority). Pass explicit arguments to override these configs')
    return parser


def to_dict(hparams):
    if hasattr(hparams, 'items'):
        return dict(hparams.items())
    return vars(hparams)


def normalize_key(k):
    return k.replace('-', '_')


def normalize_value(v):
    try:
        return int(v)
    except:
        pass
    try:
        return float(v)
    except:
        pass
    return v


def add_unknown_arguments(args, unknown, silent=False):
    args = copy.deepcopy(args)

    i = 0

    if len(unknown) > 0 and not silent:
        logger.warning(f"Parsing unkown arguments: {unknown}. Please check for typos")

    while i < len(unknown):

        arg = unknown[i]
        if not arg.startswith("-"):
            if not silent:
                logger.error(f"Cannot correctly parse: {unknown[i]} arguments as it as it does not start with \'-\' sign")
            i += 1
            continue
        if arg.startswith("--"):
            arg = arg[2:]
        else:
            arg = arg[1:]

        if arg.startswith('no-'):
            k = arg[3:]
            setattr(args, normalize_key(k), False)
            i += 1
            continue

        if '=' in arg:
            arg = arg.split('=')
            if len(arg) != 2:
                if not silent:
                    logger.error(f"Cannot correctly parse: {unknown[i]} arguments as it contains more than one \'=\' sign")
                i += 1
                continue
            k, v = arg
            setattr(args, normalize_key(k), normalize_value(v))
            i += 1
            continue

        k = normalize_key(arg)
        if i == len(unknown) - 1 or unknown[i + 1].startswith("-"):
            setattr(args, k, True)
            i += 1
        else:
            v = unknown[i + 1]
            setattr(args, k, normalize_value(v))
            i += 2

    return args


def _beam_arguments(*args, return_defaults=False, return_tags=False, silent=False, strict=False, **kwargs):
    '''
    args can be list of arguments or a long string of arguments or list of strings each contains multiple arguments
    kwargs is a dictionary of both defined and undefined arguments
    '''

    pr = args[0]
    args = args[1:]

    def update_parser(p, d):
        for pi in p._actions:
            for o in pi.option_strings:
                o = o.replace('--', '').replace('-', '_')
                if o in d:
                    p.set_defaults(**{pi.dest: d[o]})

    if is_notebook():
        sys.argv = sys.argv[:1]

    file_name = sys.argv[0] if len(sys.argv) > 0 else '/tmp/tmp.py'
    sys_args = sys.argv[1:]

    args_str = []
    args_dict = []

    for ar in args:

        ar_type = check_type(ar)

        if isinstance(ar, Namespace):
            args_dict.append(to_dict(ar))
        elif ar_type.minor == 'dict':
            args_dict.append(ar)
        elif ar_type.major == 'scalar' and ar_type.element == 'str':
            args_str.append(ar)
        else:
            raise ValueError

    for ar in args_dict:
        kwargs = {**kwargs, **ar}

    args_str = re.split(r"\s+", ' '.join([ar.strip() for ar in args_str]))

    sys.argv = [file_name] + args_str + sys_args
    sys.argv = list(filter(lambda x: bool(x), sys.argv))

    update_parser(pr, kwargs)
    # set defaults from environment variables
    update_parser(pr, os.environ)

    if return_defaults:
        args = pr.parse_args([])
    else:
        args, unknown = pr.parse_known_args()
        if not strict:
            args = add_unknown_arguments(args, unknown, silent=silent)

    for k, v in kwargs.items():
        if k not in args:
            setattr(args, k, v)

    tags = defaultdict(set)
    if hasattr(args, 'config_files') and args.config_files:
        config_files = args.config_files
        delattr(args, 'config_files')

        config_args = {}

        for config_file in config_files:
            cf = beam_path(config_file).read()
            if '_tags' in cf:
                for k, v in cf['_tags'].items():
                    tags[k].add(v)
                del cf['_tags']
            config_args.update(cf)

        args = Namespace(**{**config_args, **to_dict(args)})
    elif hasattr(args, 'config_files'):
        delattr(args, 'config_files')

    beam_key.set_hparams(to_dict(args))

    if not return_tags:
        return args

    for pai in pr._actions:
        tag_list = get_tags_from_action(pai)
        for tag in tag_list:
            tags[tag].add(pai.dest)

    return args, tags


def get_tags_from_action(action):
    tags = set()
    if action.metavar is not None:
        tags = re.findall(r'[a-zA-Z0-9_-]+', action.metavar)
    return list(tags)


def get_beam_llm(llm_uri=None, get_from_key=True):
    llm = None
    if llm_uri is None and get_from_key:
        llm_uri = beam_key('BEAM_LLM', store=False)
    if llm_uri is not None:
        try:
            from .llm import beam_llm
            llm = beam_llm(llm_uri)
        except ImportError:
            pass
    return llm


def print_beam_hyperparameters(args, default_params=None, debug_only=False):
    if debug_only:
        log_func = logger.debug
    else:
        log_func = logger.info

    log_func(f"Beam experiment (Beam version: {__version__})")
    log_func(f"project: {args.project_name}, algorithm: {args.algorithm}, identifier: {args.identifier}")
    log_func(f"Global paths:")
    log_func(f"data-path (where the dataset should be): {args.data_path}")
    log_func(f"logs-path (where the experiments are log to): {args.logs_path}")
    log_func(f'Experiment objective: {args.objective} (set for schedulers, early stopping and best checkpoint store)')
    log_func('Experiment Hyperparameters (only non default values are listed):')
    log_func('----------------------------------------------------------'
             '---------------------------------------------------------------------')

    if hasattr(args, 'hparams'):
        hparams_list = args.hparams
    else:
        hparams_list = args

    var_args_sorted = dict(sorted(to_dict(args).items()))

    for k, v in var_args_sorted.items():
        if k == 'hparams':
            continue
        elif (default_params is not None and k in hparams_list and hasattr(default_params, k) and
              (v is not None and v != getattr(default_params, k))):
            log_func(k + ': ' + str(v))
        else:
            logger.debug(k + ': ' + str(v))

    log_func('----------------------------------------------------------'
             '---------------------------------------------------------------------')



from ..core.processor import Processor
from ..utils import collate_chunks


class Reducer(Processor):

    def __init__(self, hparams, *args, dim=1, **kwargs):
        super().__init__(hparams, *args, **kwargs)
        self.dim = dim

    def reduce(self, *xs, **kwargs):
        return collate_chunks(*xs, dim=self.dim, **kwargs)



import time
from ..logger import beam_logger as logger
from ray.tune.stopper import Stopper
from typing import Union
import datetime


class TimeoutStopper(Stopper):
    """Stops all trials after a certain timeout.

    This stopper is automatically created when the `time_budget_s`
    argument is passed to `tune.run()`.

    Args:
        timeout: Either a number specifying the timeout in seconds, or
            a `datetime.timedelta` object.
    """

    def __init__(self, timeout: Union[int, float, datetime.timedelta]):
        from datetime import timedelta

        if isinstance(timeout, timedelta):
            self._timeout_seconds = timeout.total_seconds()
        elif isinstance(timeout, (int, float)):
            self._timeout_seconds = timeout
        else:
            raise ValueError(
                "`timeout` parameter has to be either a number or a "
                "`datetime.timedelta` object. Found: {}".format(type(timeout))
            )

        self._budget = self._timeout_seconds

        self.start_time = {}

    def stop_all(self):
        return False

    def __call__(self, trial_id, result):
        now = time.time()

        if trial_id in self.start_time:
            if now - self.start_time[trial_id] >= self._budget:
                logger.info(
                    f"Reached timeout of {self._timeout_seconds} seconds. "
                    f"Stopping this trials."
                )
                return True
        else:
            self.start_time[trial_id] = now

        return False


import uuid
from typing import Any, Dict, List, Union
from ..logger import beam_logger as logger


class Conversation:
    """
    Utility class containing a conversation and its history. This class is meant to be used as an input to the
    [`ConversationalPipeline`]. The conversation contains several utility functions to manage the addition of new user
    inputs and generated model responses.

    Arguments:
        messages (Union[str, List[Dict[str, str]]], *optional*):
            The initial messages to start the conversation, either a string, or a list of dicts containing "role" and
            "content" keys. If a string is passed, it is interpreted as a single message with the "user" role.
        conversation_id (`uuid.UUID`, *optional*):
            Unique identifier for the conversation. If not provided, a random UUID4 id will be assigned to the
            conversation.

    Usage:

    ```python
    conversation = Conversation("Going to the movies tonight - any suggestions?")
    conversation.add_message({"role": "assistant", "content": "The Big lebowski."})
    conversation.add_message({"role": "user", "content": "Is it good?"})
    ```"""

    def __init__(
        self, messages: Union[str, List[Dict[str, str]]] = None, conversation_id: uuid.UUID = None, **deprecated_kwargs
    ):
        if not conversation_id:
            conversation_id = uuid.uuid4()

        if messages is None:
            text = deprecated_kwargs.pop("text", None)
            if text is not None:
                messages = [{"role": "user", "content": text}]
            else:
                messages = []
        elif isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        # This block deals with the legacy args - new code should just totally
        # avoid past_user_inputs and generated_responses
        self._num_processed_user_inputs = 0
        generated_responses = deprecated_kwargs.pop("generated_responses", None)
        past_user_inputs = deprecated_kwargs.pop("past_user_inputs", None)
        if generated_responses is not None and past_user_inputs is None:
            raise ValueError("generated_responses cannot be passed without past_user_inputs!")
        if past_user_inputs is not None:
            legacy_messages = []
            if generated_responses is None:
                generated_responses = []
            # We structure it this way instead of using zip() because the lengths may differ by 1
            for i in range(max([len(past_user_inputs), len(generated_responses)])):
                if i < len(past_user_inputs):
                    legacy_messages.append({"role": "user", "content": past_user_inputs[i]})
                if i < len(generated_responses):
                    legacy_messages.append({"role": "assistant", "content": generated_responses[i]})
            messages = legacy_messages + messages

        self.uuid = conversation_id
        self.messages = messages

    def __eq__(self, other):
        if not isinstance(other, Conversation):
            return False
        return self.uuid == other.uuid or self.messages == other.messages

    def add_message(self, message: Dict[str, str]):
        if not set(message.keys()) == {"role", "content"}:
            raise ValueError("Message should contain only 'role' and 'content' keys!")
        if message["role"] not in ("user", "assistant", "system"):
            raise ValueError("Only 'user', 'assistant' and 'system' roles are supported for now!")
        self.messages.append(message)

    def add_user_input(self, text: str, overwrite: bool = False):
        """
        Add a user input to the conversation for the next round. This is a legacy method that assumes that inputs must
        alternate user/assistant/user/assistant, and so will not add multiple user messages in succession. We recommend
        just using `add_message` with role "user" instead.
        """
        if len(self) > 0 and self[-1]["role"] == "user":
            if overwrite:
                logger.warning(
                    f'User input added while unprocessed input was existing: "{self[-1]["content"]}" was overwritten '
                    f'with: "{text}".'
                )
                self[-1]["content"] = text
            else:
                logger.warning(
                    f'User input added while unprocessed input was existing: "{self[-1]["content"]}" new input '
                    f'ignored: "{text}". Set `overwrite` to True to overwrite unprocessed user input'
                )
        else:
            self.messages.append({"role": "user", "content": text})

    def append_response(self, response: str):
        """
        This is a legacy method. We recommend just using `add_message` with an appropriate role instead.
        """
        self.messages.append({"role": "assistant", "content": response})

    def mark_processed(self):
        """
        This is a legacy method, as the Conversation no longer distinguishes between processed and unprocessed user
        input. We set a counter here to keep behaviour mostly backward-compatible, but in general you should just read
        the messages directly when writing new code.
        """
        self._num_processed_user_inputs = len(self._user_messages)

    def __iter__(self):
        for message in self.messages:
            yield message

    def __getitem__(self, item):
        return self.messages[item]

    def __setitem__(self, key, value):
        self.messages[key] = value

    def __len__(self):
        return len(self.messages)

    def __repr__(self):
        """
        Generates a string representation of the conversation.

        Returns:
            `str`:

        Example:
            Conversation id: 7d15686b-dc94-49f2-9c4b-c9eac6a1f114 user: Going to the movies tonight - any suggestions?
            bot: The Big Lebowski
        """
        output = f"Conversation id: {self.uuid}\n"
        for message in self.messages:
            output += f"{message['role']}: {message['content']}\n"
        return output

    def iter_texts(self):
        # This is a legacy method for backwards compatibility. It is recommended to just directly access
        # conversation.messages instead.
        for message in self.messages:
            yield message["role"] == "user", message["content"]

    @property
    def _user_messages(self):
        # This is a legacy property for backwards compatibility. It is recommended to just directly access
        # conversation.messages instead.
        return [message["content"] for message in self.messages if message["role"] == "user"]

    @property
    def past_user_inputs(self):
        # This is a legacy property for backwards compatibility. It is recommended to just directly access
        # conversation.messages instead. The modern class does not care about which messages are "processed"
        # or not.
        if not self._user_messages:
            return []
        # In the past, the most recent user message had to be mark_processed() before being included
        # in past_user_messages. The class essentially had a single-message buffer, representing messages that
        # had not yet been replied to. This is no longer the case, but we mimic the behaviour in this property
        # for backward compatibility.
        if self.messages[-1]["role"] != "user" or self._num_processed_user_inputs == len(self._user_messages):
            return self._user_messages

        return self._user_messages[:-1]

    @property
    def generated_responses(self):
        # This is a legacy property for backwards compatibility. It is recommended to just directly access
        # conversation.messages instead.
        return [message["content"] for message in self.messages if message["role"] == "assistant"]

    @property
    def new_user_input(self):
        # This is a legacy property for backwards compatibility. It is recommended to just directly access
        # conversation.messages instead.
        return self._user_messages[-1]


from argparse import Namespace
from ..openai import simulate_openai_chat, simulate_openai_completion


openai_legacy_simulator = Namespace(ChatCompletion=Namespace(create=simulate_openai_chat),
                             Completion=Namespace(create=simulate_openai_completion))




import copy

from ..utils import find_port, check_type, is_notebook, beam_device
from ..logger import beam_logger as logger
from ..path import beam_path, BeamPath
import pandas as pd
import optuna
from functools import partial
from ..experiment import Experiment

import numpy as np
from scipy.special import erfinv
from .core import BeamHPO


class OptunaBase:

    @staticmethod
    def _linspace(trial, param, start, end, n_steps, endpoint=True,  dtype=None):
        x = np.linspace(start, end, n_steps, endpoint=endpoint)
        if np.sum(np.abs(x - np.round(x))) < 1e-8 or dtype in [int, np.int, np.int64, 'int', 'int64']:
            x = np.round(x).astype(int)
        i = trial.suggest_int(param, 0, len(x) - 1)
        return x[i]

    @staticmethod
    def _logspace(trial, param, start, end, n_steps, base=None, dtype=None):
        x = np.logspace(start, end, n_steps, base=base)
        if np.sum(np.abs(x - np.round(x))) < 1e-8 or dtype in [int, np.int, np.int64, 'int', 'int64']:
            x = np.round(x).astype(int)
        i = trial.suggest_int(param, 0, len(x) - 1)
        return x[i]

    @staticmethod
    def _uniform(trial, param, start, end):
        return trial.suggest_uniform(param, start, end)

    @staticmethod
    def _loguniform(trial, param, start, end):
        return trial.suggest_loguniform(param, start, end)

    @staticmethod
    def _categorical(trial, param, choices):
        return trial.suggest_categorical(param, choices)

    @staticmethod
    def _randn(trial, param, mu, sigma):
        x = trial.suggest_uniform(param, 0, 1)
        return mu + sigma * np.sqrt(2) * erfinv(2 * x - 1)


class OptunaHPO(BeamHPO, OptunaBase):

    def runner(self, trial, suggest):

        config = suggest(trial)
        hparams = self.generate_hparams(config)

        experiment = Experiment(hparams, hpo='optuna', trial=trial, print_hyperparameters=False)

        logger.info(f"Experiment directory is: {experiment.experiment_dir}, see experiment_dir attr in trial logs")
        trial.set_user_attr("experiment_dir", experiment.experiment_dir)

        alg, results = experiment.fit(alg=self.alg, algorithm_generator=self.ag, return_results=True)
        if self.post_train_hook is not None:
            self.post_train_hook(alg=alg, experiment=experiment, hparams=hparams, suggestion=config, results=results)

        self.tracker(algorithm=alg, results=results, hparams=hparams, suggestion=config)

        return results.objective

    def grid_search(self, load_study=False, storage=None, sampler=None, pruner=None, study_name=None, direction=None,
                    load_if_exists=False, directions=None, sync_parameters=None, explode_parameters=None, **kwargs):

        df_sync = pd.DataFrame(sync_parameters)
        df_explode = pd.DataFrame([explode_parameters])
        for c in list(df_explode.columns):
            df_explode = df_explode.explode(c)

        if sync_parameters is None:
            df = df_explode
        elif explode_parameters is None:
            df = df_sync
        else:
            df = df_sync.merge(df_explode, how='cross')

        df = df.reset_index(drop=True)
        n_trials = len(df)

        if not 'cpu' in self.device.type:
            if 'n_jobs' not in kwargs or kwargs['n_jobs'] != 1:
                logger.warning("Optuna does not support multi-GPU jobs. Setting number of parallel jobs to 1")
            kwargs['n_jobs'] = 1

        if study_name is None:
            study_name = f'{self.hparams.project_name}/{self.hparams.algorithm}/{self.hparams.identifier}'

        if direction is None:
            direction = 'maximize'

        if storage is None:
            if self.hpo_path is not None:

                path = beam_path(self.hpo_path)
                path.joinpath('optuna').mkdir(parents=True, exist_ok=True)

                storage = f'sqlite:///{self.hpo_path}/{study_name}.db'

        if load_study:
            study = optuna.create_study(storage=storage, sampler=sampler, pruner=pruner, study_name=study_name)
        else:
            study = optuna.create_study(storage=storage, sampler=sampler, pruner=pruner, study_name=study_name,
                                        direction=direction, load_if_exists=load_if_exists, directions=directions)

        for it in df.iterrows():
            study.enqueue_trial(it[1].to_dict())

        def dummy_suggest(trial):
            config = {}
            for k, v in it[1].items():
                v_type = check_type(v)
                if v_type.element == 'int':
                    config[k] = trial.suggest_int(k, 0, 1)
                elif v_type.element == 'str':
                    config[k] = trial.suggest_categorical(k, ['a', 'b'])
                else:
                    config[k] = trial.suggest_float(k, 0, 1)

            return config

        runner = partial(self.runner, suggest=dummy_suggest)
        study.optimize(runner, n_trials=n_trials, **kwargs)

        return study

    def run(self, suggest=None, load_study=False, storage=None, sampler=None, pruner=None, study_name=None,
            direction=None, load_if_exists=False, directions=None, *args, **kwargs):

        if suggest is None:
            suggest = self.get_suggestions

        if not 'cpu' in self.device.type:
            if 'n_jobs' not in kwargs or kwargs['n_jobs'] != 1:
                logger.warning("Optuna does not support multi-GPU jobs. Setting number of parallel jobs to 1")
            kwargs['n_jobs'] = 1

        if direction is None:
            direction = 'maximize'

        if study_name is None:
            study_name = f'{self.hparams.project_name}/{self.hparams.algorithm}/{self.hparams.identifier}'

        if storage is None:
            if self.hpo_path is not None:

                path = beam_path(self.hpo_path)
                path.joinpath('optuna').mkdir(parents=True, exist_ok=True)

                storage = f'sqlite:///{self.hpo_path}/{study_name}.db'
                logger.info(f"Using {storage} as storage to store the trials results")

        runner = partial(self.runner, suggest=suggest)

        if load_study:
            study = optuna.load_study(storage=storage, sampler=sampler, pruner=pruner, study_name=study_name)
        else:
            study = optuna.create_study(storage=storage, sampler=sampler, pruner=pruner, study_name=study_name,
                                        direction=direction, load_if_exists=load_if_exists, directions=directions)

        study.optimize(runner, *args, gc_after_trial=True, **kwargs)

        return study



import ray

from ..utils import get_class_properties, lazy_property
from ..core import Processor
from ..path import BeamURL
from ..core import MetaAsyncResult, MetaDispatcher


class RayAsyncResult(MetaAsyncResult):

    @classmethod
    def from_str(cls, value, **kwargs):
        return cls(ray.ObjectRef(bytes.fromhex(value)))

    @property
    def value(self):
        if self._value is None:
            self._value = ray.get(self.obj)
        return self._value

    def wait(self, timeout=None):
        # ready, not_ready = ray.wait([self.obj], num_returns=1, timeout=timeout)
        # return ready, not_ready
        try:
            res = ray.get(self.obj, timeout=timeout)
            return res
        except ray.exceptions.GetTimeoutError:
            return None

    @property
    def hex(self):
        return self.obj.hex()

    @property
    def is_ready(self):
        if not self._is_ready:
            ready = self.wait(timeout=0)
            self._is_ready = ready is not None
        return self._is_ready

    def __repr__(self):
        return f"AsyncResult({self.str}, is_ready={self.is_ready}, is_success={self.is_success})"

    @property
    def state(self):
        if self.is_ready:
            return "SUCCESS" if self.is_success else "FAILURE"
        else:
            return "PENDING"


class RayCluster(Processor):

    def __init__(self, *args, name=None, address=None, host=None, port=None,
                    username=None, password=None, ray_kwargs=None, **kwargs):

        super().__init__(*args, name=name, **kwargs)

        if address is None:
            if host is None and port is None:
                address = 'auto'
            else:
                if host is None:
                    host = 'localhost'
                address = BeamURL(host=host, port=port, username=username, password=password)
                address = address.url

        ray_kwargs = ray_kwargs if ray_kwargs is not None else {}
        self.init_ray(address=address, ignore_reinit_error=True, **ray_kwargs)

    def wait(self, results, num_returns=1, timeout=None):
        results = [r.result if isinstance(r, RayAsyncResult) else r for r in results]
        return ray.wait(results, num_returns=num_returns, timeout=timeout)

    @staticmethod
    def init_ray(address=None, num_cpus=None, num_gpus=None, resources=None, labels=None, object_store_memory=None,
                 ignore_reinit_error=False, include_dashboard=True, dashboard_host='0.0.0.0',
                 dashboard_port=None, job_config=None, configure_logging=True, logging_level=None, logging_format=None,
                 log_to_driver=True, namespace=None, runtime_env=None, storage=None):
        kwargs = {}
        if logging_level is not None:
            kwargs['logging_level'] = logging_level

        if not ray.is_initialized():
            ray.init(address=address, num_cpus=num_cpus, num_gpus=num_gpus, resources=resources, labels=labels,
                     object_store_memory=object_store_memory, ignore_reinit_error=ignore_reinit_error,
                     job_config=job_config, configure_logging=configure_logging, logging_format=logging_format,
                     log_to_driver=log_to_driver, namespace=namespace, storage=storage,
                     runtime_env=runtime_env, dashboard_port=dashboard_port,
                     include_dashboard=include_dashboard, dashboard_host=dashboard_host, **kwargs)

    @staticmethod
    def shutdown():
        ray.shutdown()

    @property
    def head_node_ip(self):
        return ray.util.get_node_ip_address()


class RayRemoteClass:

    def __init__(self, remote_class, asynchronous=False, properties=None):
        self.remote_class = remote_class
        self.asynchronous = asynchronous
        self.properties = properties if properties is not None else []

    def remote_wrapper(self, method):
        def wrapper(*args, **kwargs):
            res = method.remote(*args, **kwargs)
            if self.asynchronous:
                return RayAsyncResult(res)
            else:
                return ray.get(res)
        return wrapper

    def kill(self, no_restart=False):
        ray.kill(self.remote_class, no_restart=no_restart)

    def __getattr__(self, item):

        if item in self.properties:
            res = self.remote_class._get_property.remote(item)
            if self.asynchronous:
                return RayAsyncResult(res)
            else:
                return ray.get(res)

        return self.remote_wrapper(getattr(self.remote_class, item))

    def __call__(self, *args, **kwargs):
        res = self.remote_class.__call__.remote(*args, **kwargs)
        if self.asynchronous:
            return RayAsyncResult(res)
        else:
            return ray.get(res)


class RayDispatcher(MetaDispatcher, RayCluster):

    def __init__(self, obj, *routes, name=None, address=None, host=None, port=None,
                 username=None, password=None, remote_kwargs=None, ray_kwargs=None, asynchronous=True, **kwargs):

        MetaDispatcher.__init__(self, obj, *routes, name=name, ray_kwargs=ray_kwargs,
                                asynchronous=asynchronous, **kwargs)
        RayCluster.__init__(self, name=name, address=address, host=host, port=port, username=username,
                            password=password, ray_kwargs=ray_kwargs, **kwargs)

        self.remote_kwargs = remote_kwargs if remote_kwargs is not None else {}

        if self.type == 'function':
            self.call_function = self.remote_function_wrapper(self.obj)
        elif self.type == 'instance':
            if hasattr(self.obj, '__call__'):
                self.call_function = self.remote_function_wrapper(self.remote_method_wrapper(self.obj,
                                                                                             '__call__'))
            for route in self.routes:
                if hasattr(self.obj, route):
                    self._routes_methods[route] = self.remote_function_wrapper(
                        self.remote_method_wrapper(self.obj, route))

        elif self.type == 'class':
            self.call_function = self.remote_class_wrapper(self.obj)
        else:
            raise ValueError(f"Unknown type: {self.type}")

    @lazy_property
    def route_methods(self):
        return self._routes_methods

    def poll(self, task_id, timeout=0):
        async_res = RayAsyncResult.from_str(task_id)
        return async_res.wait(timeout=timeout)

    @property
    def ray_remote(self):
        return ray.remote(**self.remote_kwargs) if len(self.remote_kwargs) else ray.remote

    def remote_method_wrapper(self, obj, method_name):
        method = getattr(obj, method_name)

        def wrapper(*args, **kwargs):
            res = method(*args, **kwargs)
            return res

        return wrapper

    def remote_class_wrapper(self, cls):

        @self.ray_remote
        class RemoteClassWrapper(cls):
            def _get_property(self, prop):
                return getattr(self, prop)

        def wrapper(*args, **kwargs):
            res = RemoteClassWrapper.remote(*args, **kwargs)
            return res

        return wrapper

    def remote_function_wrapper(self, func, asynchronous=None, bypass=False):

        if asynchronous is None:
            asynchronous = self.asynchronous

        func = self.ray_remote(func)

        def wrapper(*args, **kwargs):
            res = func.remote(*args, **kwargs)
            if bypass:
                return res
            elif asynchronous:
                return RayAsyncResult(res)
            else:
                return ray.get(res)
        return wrapper

    def __call__(self, *args, **kwargs):
        assert self.call_function is not None, "No function to call"
        res = self.call_function(*args, **kwargs)
        if self.type == 'class':
            properties = get_class_properties(self.obj)
            return RayRemoteClass(res, asynchronous=self.asynchronous, properties=properties)
        else:
            return res



import numpy as np

import scipy.sparse as sp
from .core import BeamSimilarity, Similarities


class SparnnSimilarity(BeamSimilarity):

    def __init__(self, *args, k_clusters=10, matrix_size=None, num_indexes=2, **kwargs):
        super().__init__(*args, k_clusters=k_clusters, matrix_size=matrix_size, num_indexes=num_indexes, **kwargs)
        self.k_clusters = self.get_hparam('k_clusters', k_clusters)
        self.matrix_size = self.get_hparam('matrix_size', matrix_size)
        self.num_indexes = self.get_hparam('num_indexes', num_indexes)

        self.index = None
        self.vectors = None
        self.cluster = None

    def reset(self):
        self.index = None
        self.vectors = None
        self.cluster = None

    @property
    def exclude_pickle_attributes(self):
        return ['index', 'vectors', 'cluster']

    def add(self, x, index=None, **kwargs):

        x, index = self.extract_data_and_index(x, index, convert_to='scipy_csr')
        self.add_index(x, index)
        if self.vectors is None:
            self.vectors = x
        else:
            self.vectors = sp.vstack([self.vectors, x])

    def fit(self, x=None, index=None, **kwargs):

        if x is not None:
            self.add(x, index)
        if self.vectors is None:
            raise ValueError('No vectors to fit')

        import pysparnn.cluster_index as ci
        self.cluster = ci.MultiClusterIndex(self.vectors, np.arange(len(self.index)), num_indexes=self.num_indexes,
                                            matrix_size=self.matrix_size)

    def search(self, x, k=1):
        x, _ = self.extract_data_and_index(x, convert_to='scipy_csr')
        if self.cluster is None:
            self.fit()

        res = self.cluster.search(x, k=k, k_clusters=self.k_clusters, return_distance=True)
        res = np.array(res).transpose(2, 0, 1)
        I = res[1].astype(np.int64)
        D = res[0]

        return Similarities(index=self.index[I], distance=D, metric='cosine', model='sparnn')



import os
from collections import defaultdict
from .core import BeamConfig, BeamParam
from ..path import beam_path


def deepspeed_dtype_mapper(dtype):
    model_mapping = {'float32': 'fp32',
                     'float16': 'fp16',
                     'bfloat16': 'bf16'}
    return model_mapping[dtype]


class DeepspeedConfig(BeamConfig):

    parameters = [
        BeamParam('deepspeed_config', str, None, 'Deepspeed configuration JSON file.'),

        # Optimizer Parameters

        BeamParam('deepspeed_optimizer', str, 'AdamW',
                  'Optimizer type (currently used for deepspeed configuration only) '
                  'Supported optimizers: [Adam, AdamW, Lamb, OneBitAdam, OneBitLamb]'),

        # Scheduler Parameters

        # Automatic mixed precision (AMP) training options

        # ZeRO Optimizations for FP16 Training

        BeamParam('zero_stage', int, 2, 'The ZeRO training stage to use.'),
        BeamParam('stage3_gather_16bit_weights_on_model_save', bool, False,
                  'Whether to gather 16-bit weights on model save in ZeRO stage 3'),

        # Parameter offloading

        BeamParam('offload_param_device', str, None, 'Whether to offload parameters from GPU in ZeRO stage 3: '
                                                     '[cpu, nvme, none]'),
        BeamParam('offload_param_pin_memory', bool, True, 'Whether to pin memory for offloaded parameters'),

        BeamParam('offload_param_nvme_path', str, os.path.join(os.path.expanduser('~'), 'beam_projects', 'deepspeed'),
                  'Path to NVMe device for offloaded parameters'),

        # Optimizer offloading

        BeamParam('offload_optimizer_device', str, None,
                  'Whether to offload optimizer states from GPU in ZeRO stages 1/2/3: '
                  '[cpu, nvme, none]'),
        BeamParam('offload_optimizer_pin_memory', bool, True, 'Whether to pin memory for offloaded optimizer states'),

        BeamParam('autotuning', bool, False, 'Whether to use deepspeed autotuning feature.'),

        # Activation Checkpointing

        BeamParam('partition_activations', bool, False,
                  'Enables partition activation when used with model parallelism'),
        BeamParam('cpu_checkpointing', bool, False,
                  'Offloads partitioned activations to CPU if partition_activations is enabled'),
        BeamParam('contiguous_memory_optimization', bool, False,
                  'Copies partitioned activations so that they are contiguous in memory'),
        BeamParam('number_checkpoints', int, None,
                  'Total number of activation checkpoints used to allocate memory buffer '
                  'for contiguous_memory_optimization'),
        BeamParam('synchronize_checkpoint_boundary', bool, False,
                  'Inserts get_accelerator().synchronize() at each checkpoint boundary'),
        BeamParam('profile', bool, False, 'Logs the forward and backward time for each checkpoint function'),

        # Sparse Attention

        # Data Efficiency

        # Data Type options
        BeamParam('grad_accum_dtype', str, None, 'The data type for gradient accumulation.'
                                                 'Supported types: [float32, float16, bfloat16]'),
    ]


def recursive_dict_update(d, u):
    '''

    Merge two dicts recursively and update the values of the first one with the values of the second one.
    @param d:
    @param u:
    @return:
    '''
    for k, v in u.items():
        if isinstance(v, dict):
            d[k] = recursive_dict_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def recursive_to_dict(obj):
    if isinstance(obj, defaultdict) or isinstance(obj, dict):
        return dict([(k, recursive_to_dict(v)) for k, v in obj.items()])
    return obj


def deepspeed_config_generator(hparams):

    config = defaultdict(dict)
    target = hparams.get('training_framework', 'deepspeed')

    config["train_micro_batch_size_per_gpu"] = hparams.get('batch_size_train') or hparams.get('batch_size')
    config["gradient_accumulation_steps"] = hparams.get('accumulate')

    if hparams.get('zero_stage', 2) is not None:
        config['zero_optimization']['stage'] = hparams.get('zero_stage', 2)
        config['stage3_gather_16bit_weights_on_model_save'] = (
            hparams.get('stage3_gather_16bit_weights_on_model_save', False))

        if hparams.get('offload_param_device', None) is not None:
            config['zero_optimization']['offload_param'] = {}
            config['zero_optimization']['offload_param']['device'] = hparams.get('offload_param_device')
            config['zero_optimization']['offload_param']['pin_memory'] = hparams.get('offload_param_pin_memory', False)

        if hparams.get('offload_optimizer_device', None) is not None:
            config['zero_optimization']['offload_optimizer'] = {}
            config['zero_optimization']['offload_optimizer']['device'] = hparams.get('offload_optimizer_device')
            config['zero_optimization']['offload_optimizer']['pin_memory'] = hparams.get('offload_optimizer_pin_memory', False)

    # optimizer
    if target == 'deepspeed':
        config['optimizer']['type'] = hparams.get('deepspeed_optimizer', 'AdamW')
        config['optimizer']['params'] = {'lr': hparams.get('lr-dense')}
        if hparams.get('weight_decay', None) is not None:
            config['optimizer']['params']['weight_decay'] = hparams.get('weight_decay')
        if 'adam' in config['optimizer']['type'].lower():
            config['optimizer']['params']['betas'] = [hparams.get('momentum', 0.8), hparams.get('beta2', 0.999)]
            config['optimizer']['params']['eps'] = hparams.get('eps', 1e-8)

    # activation_checkpointing
    config['activation_checkpointing']['partition_activations'] = (
        hparams.get('partition_activations', False))
    config['activation_checkpointing']['cpu_checkpointing'] = hparams.get('cpu_checkpointing', False)
    config['activation_checkpointing']['contiguous_memory_optimization'] = (
        hparams.get('contiguous_memory_optimization', False))
    config['activation_checkpointing']['number_checkpoints'] = hparams.get('number_checkpoints', None)
    config['activation_checkpointing']['synchronize_checkpoint_boundary'] = (
        hparams.get('synchronize_checkpoint_boundary', False))
    config['activation_checkpointing']['profile'] = hparams.get('profile', False)

    # autotuning
    if hparams.get('autotuning', False):
        config['autotuning']['enabled'] = True

    model_dtype = deepspeed_dtype_mapper(hparams.get('model_dtype', 'float32'))

    config['fp16']['enabled'] = model_dtype == 'fp16'
    config['bf16']['enabled'] = model_dtype == 'bf16'

    # steps_per_print
    if target != 'accelerate':
        epoch_length_train = hparams.get('epoch_length_train', None)
        epoch_length_eval = hparams.get('epoch_length_eval', None)
        if epoch_length_train is not None and epoch_length_eval is not None:
            config['steps_per_print'] = epoch_length_train + epoch_length_eval

    # update config with deepspeed_config (deepspeed_config overrides all other parameters)
    hparams_dict = recursive_to_dict(config)

    if hparams.get('deepspeed_config', None) is not None:
        config_file_dict = beam_path(hparams.get('deepspeed_config')).read()
        recursive_dict_update(hparams_dict, config_file_dict)

    return hparams_dict



from typing import List

from sentence_transformers import SentenceTransformer

from ..utils import beam_device
from ..core import Processor
from ..llm import default_tokenizer
from .dense import DenseSimilarity
from ..path import local_copy


class TextSimilarity(DenseSimilarity):
    def __init__(self, *args, dense_model_path="BAAI/bge-base-en-v1.5", tokenizer_path=None,
                 use_dense_model_tokenizer=True, dense_model=None, tokenizer=None, cache_folder=None,
                 device="cuda", batch_size=32, show_progress_bar=True, **kwargs):
        """
        Initialize the RAG (Retrieval-Augmented Generation) retriever.

        Parameters:
        data_train (pd.DataFrame): A dataframe containing the training data with a 'text' column.
        alfa (float): Weighting factor for combining dense and sparse retrieval scores.
        embedding_model (str): The name of the sentence transformer model used for embedding.
        model (str): The name of the transformer model used for causal language modeling.
        device (str): The device to run the models on (e.g., 'cuda:1' for GPU).
        """

        Processor.__init__(self, *args, device=device, tokenizer_path=tokenizer_path,
                           dense_model_path=dense_model_path, use_dense_model_tokenizer=use_dense_model_tokenizer,
                           cache_folder=cache_folder, batch_size=batch_size, show_progress_bar=show_progress_bar,
                           **kwargs)

        # Device to run the model
        self.device = beam_device(self.get_hparam('device'))
        # Load the sentence transformer model for embeddings
        self.cache_folder = self.get_hparam('cache_folder', cache_folder)
        self.batch_size = self.get_hparam('batch_size', batch_size)
        self.show_progress_bar = self.get_hparam('show_progress_bar', show_progress_bar)

        if dense_model is None:
            self.dense_model = SentenceTransformer(self.get_hparam('dense_model_path'), device=str(self.device))
        else:
            self.dense_model = dense_model
            self.dense_model.to(self.device)

        d = self.dense_model.get_sentence_embedding_dimension()

        super().__init__(*args, device=device, tokenizer_path=tokenizer_path,
                         dense_model_path=dense_model_path, use_dense_model_tokenizer=use_dense_model_tokenizer,
                         d=d, **kwargs)

        self._tokenizer = None
        if tokenizer is None:
            if self.get_hparam('tokenizer_path') is not None:
                from transformers import PreTrainedTokenizerFast
                self._tokenizer = PreTrainedTokenizerFast(tokenizer_file=self.get_hparam('tokenizer_path'))
        else:
            self._tokenizer = tokenizer

    @property
    def tokenizer(self):

        tokenizer = self._tokenizer
        if self._tokenizer is None:
            tokenizer = default_tokenizer
            if self.get_hparam('use_dense_model_tokenizer'):
                tokenizer = self.dense_model.tokenize

        return tokenizer

    def add(self, x, index=None, **kwargs):

        x, index = self.extract_data_and_index(x, index)
        dense_vectors = self.encode(x)
        super().add(dense_vectors, index)

    def encode(self, x: List[str]):
        x = list(x)
        return self.dense_model.encode(x, batch_size=self.batch_size, show_progress_bar=True, convert_to_tensor=True)

    def search(self, x: List[str], k=1):

        x, _ = self.extract_data_and_index(x)
        dense_vectors = self.encode(x)
        similarities = super().search(dense_vectors, k)
        return similarities

    @property
    def exclude_pickle_attributes(self):
        return super().exclude_pickle_attributes + ['dense_model', '_tokenizer']

    def save_state(self, path, ext=None):

        super().save_state(path, ext)

        if self._tokenizer is not None:
            if hasattr(self._tokenizer, 'save_pretrained'):
                tokenizer_path = path.joinpath('tokenizer.hf')
                with local_copy(tokenizer_path) as p:
                    self._tokenizer.save_pretrained(p)
            else:
                tokenizer_path = path.joinpath('tokenizer.pkl')
                tokenizer_path.write(self._tokenizer)

    def load_state(self, path):

        super().load_state(path)
        self.dense_model = SentenceTransformer(self.get_hparam('dense_model_path'), device=str(self.device))
        if path.joinpath('tokenizer.hf').exists():
            from transformers import PreTrainedTokenizerFast
            with local_copy(path.joinpath('tokenizer.hf')) as p:
                self._tokenizer = PreTrainedTokenizerFast(tokenizer_file=p)
        elif path.joinpath('tokenizer.pkl').exists():
            self._tokenizer = path.joinpath('tokenizer.pkl').read()




import inspect
import ast
import sys

from ..core import Processor
from .utils import get_module_paths, ImportCollector, is_installed_package, is_std_lib, get_origin, is_module_installed
from ..utils import lazy_property
from ..path import beam_path

import importlib.metadata

import pkg_resources
import os
import importlib
import warnings

from ..logger import beam_logger as logger
from uuid import uuid4 as uuid


class AutoBeam(Processor):

    # Blacklisted pip packages (sklearn is a fake project that should be ignored, scikit-learn is the real one)
    blacklisted_pip_package = ['sklearn']

    def __init__(self, obj):
        self._private_modules = None
        self._visited_modules = None
        self.obj = obj

    @lazy_property
    def self_path(self):
        return beam_path(inspect.getfile(AutoBeam)).resolve()

    @lazy_property
    def loaded_modules(self):
        modules = list(sys.modules.keys())
        root_modules = [m.split('.')[0] for m in modules]
        return set(root_modules).union(set(modules))

    @property
    def visited_modules(self):
        if self._visited_modules is None:
            self._visited_modules = set()
        return self._visited_modules

    @property
    def private_modules(self):
        if self._private_modules is None:
            self._private_modules = [self.module_spec]
            _ = self.module_dependencies
        return self._private_modules

    def add_private_module_spec(self, module_name):
        if self._private_modules is None:
            self._private_modules = [self.module_spec]

        module_spec = importlib.util.find_spec(module_name)
        if module_spec is None:
            return
        self._private_modules.append(module_spec)

    @lazy_property
    def module_spec(self):
        module_spec = importlib.util.find_spec(type(self.obj).__module__)
        root_module = module_spec.name.split('.')[0]
        return  importlib.util.find_spec(root_module)

    @staticmethod
    def module_walk(root_path):

        root_path = beam_path(root_path).resolve()
        module_walk = {}
        # if root_path.is_file():
        #     module_walk = {'..': {root_path.name: root_path.read()}}
        #     return module_walk

        for r, dirs, files in root_path.walk():

            r_relative = r.relative_to(root_path)
            dir_files = {}
            for f in files:
                p = r.joinpath(f)
                if p.suffix == '.py':
                    dir_files[f] = p.read()
            if len(dir_files):
                module_walk[str(r_relative)] = dir_files

        return module_walk

    @lazy_property
    def private_modules_walk(self):

        private_modules_walk = {}
        root_paths = set(sum([get_module_paths(m) for m in self.private_modules if m is not None], []))
        for root_path in root_paths:
            private_modules_walk[root_path] = self.module_walk(root_path)

        return private_modules_walk

    def recursive_module_dependencies(self, module_path):

        if module_path is None:
            return set()
        module_path = beam_path(module_path).resolve()
        if str(module_path) in self.visited_modules:
            return set()
        else:
            self.visited_modules.add(str(module_path))

        try:
            content = module_path.read()
        except:
            logger.warning(f"Could not read module: {module_path}")
            return set()

        ast_tree = ast.parse(content)
        collector = ImportCollector()
        collector.visit(ast_tree)
        import_nodes = collector.import_nodes

        modules = set()
        for a in import_nodes:
            if type(a) is ast.Import:
                for ai in a.names:
                    root_name = ai.name.split('.')[0]

                    if is_installed_package(root_name) and not is_std_lib(root_name):
                        if root_name in self.loaded_modules:
                            modules.add(root_name)
                    elif not is_installed_package(root_name) and not is_std_lib(root_name):
                        if root_name in ['__main__']:
                            continue
                        try:
                            self.add_private_module_spec(root_name)
                            path = beam_path(get_origin(ai))
                            if path in [module_path, self.self_path, None]:
                                continue
                        except ValueError:
                            logger.warning(f"Could not find module: {root_name}")
                            continue
                        internal_modules = self.recursive_module_dependencies(path)
                        modules = modules.union(internal_modules)

            elif type(a) is ast.ImportFrom:

                root_name = a.module.split('.')[0]

                if a.level == 0 and (not is_std_lib(root_name)) and is_installed_package(root_name):
                    if root_name in self.loaded_modules:
                        modules.add(root_name)
                elif not is_installed_package(root_name) and not is_std_lib(root_name):
                    if a.level == 0:

                        self.add_private_module_spec(root_name)

                        path = beam_path(get_origin(a.module))
                        if path in [module_path, self.self_path, None]:
                            continue
                        internal_modules = self.recursive_module_dependencies(path)
                        modules = modules.union(internal_modules)

                    else:

                        path = module_path
                        for i in range(a.level):
                            path = path.parent

                        path = path.joinpath(f"{a.module.replace('.', os.sep)}")
                        if path.is_dir():
                            path = path.joinpath('__init__.py')
                        else:
                            path = path.with_suffix('.py')

                        internal_modules = self.recursive_module_dependencies(path)
                        modules = modules.union(internal_modules)

        return modules

    @lazy_property
    def module_dependencies(self):
        module_path = beam_path(inspect.getfile(type(self.obj))).resolve()
        modules = self.recursive_module_dependencies(module_path)
        return list(set(modules))


    @lazy_property
    def top_levels(self):

        top_levels = {}
        for i, dist in enumerate(pkg_resources.working_set):
            egg_info = beam_path(dist.egg_info).resolve()
            tp_file = egg_info.joinpath('top_level.txt')
            module_name = None
            project_name = dist.project_name

            if egg_info.parent.joinpath(project_name).is_dir():
                module_name = project_name
            elif egg_info.parent.joinpath(project_name.replace('-', '_')).is_dir():
                module_name = project_name.replace('-', '_')
            elif egg_info.joinpath('RECORD').is_file():

                record = egg_info.joinpath('RECORD').read(ext='.txt', readlines=True)
                for line in record:
                    if '__init__.py' in line:
                        module_name = line.split('/')[0]
                        break
            if module_name is None and tp_file.is_file():
                module_names = tp_file.read(ext='.txt', readlines=True)
                module_names = list(filter(lambda x: len(x) >= 2 and (not x.startswith('_')), module_names))
                if len(module_names):
                    module_name = module_names[0].strip()

            if module_name is None and egg_info.parent.joinpath(f"{project_name.replace('-', '_')}.py").is_file():
                module_name = project_name.replace('-', '_')

            if module_name is None:
                # warnings.warn(f"Could not find top level module for package: {project_name}")
                top_levels[module_name] = project_name
            elif not (module_name):
                warnings.warn(f"{project_name}: is empty")
            else:
                if module_name in top_levels:
                    if type(top_levels[module_name]) is list:
                        v = top_levels[module_name]
                    else:
                        v = [top_levels[module_name]]
                        v.append(dist)
                    top_levels[module_name] = v
                else:
                    top_levels[module_name] = dist

        return top_levels

    @property
    def import_statement(self):
        module_name = type(self.obj).__module__
        # origin = beam_path(get_origin(module_name))
        # if origin.parent.joinpath('__init__.py').is_file():
        #     module_name = f"{origin.parent.name}.{module_name}"
        class_name = type(self.obj).__name__
        return f"from {module_name} import {class_name}"

    @property
    def metadata(self):
        return {'name': self.obj.name, 'type': type(self.obj).__name__,
                'import_statement': self.import_statement, 'module_name': type(self.obj).__module__}

    @staticmethod
    def to_bundle(obj, path=None):

        if path is None:
            path = beam_path('..')
            if hasattr(obj, 'name'):
                path = path.joinpath(obj.name)
            else:
                path = path.joinpath('beam_bundle')
        else:
            path = beam_path(path)

        path = path.resolve()

        ab = AutoBeam(obj)
        path.clean()
        path.mkdir()
        logger.info(f"Saving object's files to path {path}: [requirements.json, modules.tar.gz, state, requierements.txt]")
        path.joinpath('requirements.json').write(ab.requirements)
        ab.write_requirements(ab.requirements, path.joinpath('requirements.txt'))
        ab.modules_to_tar(path.joinpath('modules.tar.gz'))
        path.joinpath('metadata.json').write(ab.metadata)
        if hasattr(obj, 'save_state'):
            obj.save_state(path.joinpath('state'))
        else:
            logger.warning(f"Object {obj} does not have a save_state method.")
        try:
            path.joinpath('skeleton.pkl').write(obj)
        except Exception as e:
            path.joinpath('skeleton.pkl').unlink()
            logger.warning(f"Could not pickle object: {obj} ({e}), saving only the state.")

        if not path.joinpath('skeleton.pkl').is_file() and \
                not path.joinpath('state').exists():
            logger.error(f"Could not save object {obj} to path {path}. "
                         f"Make sure the object has a save_state method and/or "
                         f"it is pickleable.")

        return path

    @classmethod
    def from_bundle(cls, path, cache_path=None):

        logger.info(f"Loading object from path {path}")
        if cache_path is None:
            cache_path = beam_path('/tmp/autobeam').joinpath(uuid())
        else:
            cache_path = beam_path(cache_path)

        import tarfile
        path = beam_path(path).resolve()

        # 1. Check necessary files
        req_file = path.joinpath('requirements.json')
        modules_tar = path.joinpath('modules.tar.gz')
        state_file = path.joinpath('state')
        skeleton_file = path.joinpath('skeleton.pkl')
        metadata_file = path.joinpath('metadata.json')

        if not all([file.exists() for file in [req_file, modules_tar, metadata_file]]) or \
           not any([file.exists() for file in [state_file, skeleton_file]]):
            raise ValueError(f"Path {path} does not contain all necessary files for reconstruction.")

        def load_obj():

            # 3. Extract the Python modules
            cache_path.mkdir(parents=True, exist_ok=True)
            with tarfile.open(modules_tar, "r:gz") as tar:
                tar.extractall(str(cache_path))

            # 4. Add the directory containing the extracted Python modules to sys.path
            sys.path.insert(0, str(cache_path))

            # 5. Load metadata and import necessary modules
            metadata = metadata_file.read()

            imported_class = metadata['type']
            module = importlib.import_module(metadata['module_name'])
            cls_obj = getattr(module, imported_class)

            # import_statement = metadata['import_statement']
            # exec(import_statement, globals())
            # cls_obj = globals()[imported_class]

            # 7. Construct the object from its state using a hypothetical from_state method
            if skeleton_file.is_file():
                try:
                    obj = skeleton_file.read()
                    if state_file.exists:
                        obj.load_state(state_file)
                except:
                    obj = cls_obj.from_state_path(state_file)
            else:
                obj = cls_obj.from_state_path(state_file)

            return obj

        try:
            obj = load_obj()
        except ImportError:
            logger.error(f"ImportError, some of the packages are not installed. "
                         f"Trying to install only the missing requirements.")
            # 2. Install necessary packages
            requirements = req_file.read()
            for r in requirements:
                if not is_module_installed(r['module_name']):
                    os.system(f"pip install {r['pip_package']}=={r['version']}")
            try:
                obj = load_obj()
            except Exception as e:
                logger.error(f"Exception: {e}. Trying to install all requirements.")
                all_reqs = ' '.join([f"{r['pip_package']}=={r['version']}" for r in requirements])
                os.system(f"pip install {all_reqs}")
                obj = load_obj()

        return obj

    def get_pip_package(self, module_name):

        if module_name not in self.top_levels:
            return None
        return self.top_levels[module_name]

    @lazy_property
    def requirements(self):
        requirements = []
        for module_name in self.module_dependencies:
            pip_package = self.get_pip_package(module_name)
            if pip_package is not None:
                if type(pip_package) is not list:
                    pip_package = [pip_package]
                for pp in pip_package:
                    if pp.project_name is AutoBeam.blacklisted_pip_package:
                        continue
                    requirements.append({'pip_package': pp.project_name, 'module_name': module_name,
                                         'version': pp.version
                                         })
            else:
                logger.warning(f"Could not find pip package for module: {module_name}")
        return requirements

    @staticmethod
    def write_requirements(requirements, path, relation='~=', sim_type='major'):
        '''

        @param requirements:
        @param path:
        @param relation: can be '~=', '==' or '>=' or 'all'
        @return:
        '''

        path = beam_path(path)
        if relation == 'all':
            content = '\n'.join([f"{r['pip_package']}" for r in requirements])
        elif relation in ['==', '>=']:
            content = '\n'.join([f"{r['pip_package']}{relation}{r['version']}" for r in requirements])
        elif relation == '~=':
            if sim_type == 'major':
                content = '\n'.join([f"{r['pip_package']}{relation}{'.'.join(r['version'].split('.'))[:2]}" for r in requirements])
            elif sim_type == 'minor':
                content = '\n'.join([f"{r['pip_package']}{relation}{'.'.join(r['version'].split('.'))[:3]}" for r in requirements])
            else:
                raise ValueError(f"sim_type can be 'major' or 'minor'")
        else:
            raise ValueError(f"relation can be '~=', '==' or '>=' or 'all'")

        content = f"{content}\n"
        path.write(content, ext='.txt')

    def modules_to_tar(self, path):
        path = beam_path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        import tarfile
        with tarfile.open(str(path), "w:gz") as tar:
            for i, (root_path, sub_paths) in enumerate(self.private_modules_walk.items()):
                root_path = beam_path(root_path)
                for sub_path, files in sub_paths.items():
                    for file_name, _ in files.items():
                        local_name = root_path.joinpath(sub_path, file_name)
                        relative_name = local_name.relative_to(root_path.parent)
                        tar.add(str(local_name), arcname=str(relative_name))

    @staticmethod
    def _build_image(image_name, base_image, entry_point, path):
        path = beam_path(path)
        import docker
        from docker.errors import BuildError
        # Initialize Docker client
        client = docker.from_env()

        # Define build arguments
        build_args = {
            'BASE_IMAGE': base_image,
            'REQUIREMENTS_FILE': path.joinpath('requirements.txt').str,
            'ALGORITHM_DIR': path.str,
            'ENTRYPOINT_SCRIPT': entry_point
        }

        # Path to the directory containing the Dockerfile
        path_to_dockerfile = beam_path(__file__).parent.joinpath('Dockerfile').str

        try:
            # Build the image
            image, build_logs = client.images.build(path=path_to_dockerfile, buildargs=build_args,
                                                    tag=image_name)

            # Print build logs (optional)
            for line in build_logs:
                if 'stream' in line:
                    print(line['stream'].strip())

        except BuildError as e:
            print("Error building Docker image:", e)
        except Exception as e:
            print("Error:", e)
        finally:
            client.close()



import inspect

from .processor import Processor
from ..utils import lazy_property


class MetaAsyncResult:

    def __init__(self, obj):
        self.obj = obj
        self._value = None
        self._is_ready = None
        self._is_success = None

    @classmethod
    def from_str(cls, value, **kwargs):
        raise NotImplementedError

    @property
    def value(self):
        raise NotImplementedError

    @property
    def get(self):
        return self.value

    def wait(self, timeout=None):
        raise NotImplementedError

    @property
    def hex(self):
        raise NotImplementedError

    @property
    def str(self):
        return self.hex

    @property
    def is_ready(self):
        raise NotImplementedError

    @property
    def is_success(self):
        if self._is_success is None:
            try:
                if not self.is_ready:
                    return None
                _ = self.value
                self._is_success = True
            except Exception:
                self._is_success = False
        return self._is_success

    def __str__(self):
        return self.str

    def __repr__(self):
        raise NotImplementedError

    @property
    def state(self):
        raise NotImplementedError

    @property
    def args(self):
        return None

    @property
    def kwargs(self):
        return None


class MetaDispatcher(Processor):

    def __init__(self, obj, *routes, name=None, asynchronous=True, **kwargs):

        super().__init__(name=name, **kwargs)
        self.obj = obj
        self._routes = routes
        self.asynchronous = asynchronous
        self.call_function = None
        self._routes_methods = {}

    @property
    def real_object(self):
        return self.obj

    @property
    def routes(self):
        routes = self._routes
        if routes is None or len(routes) == 0:
            routes = [name for name, attr in inspect.getmembers(self.real_object)
                      if type(name) is str and not name.startswith('_') and
                      (inspect.ismethod(attr) or inspect.isfunction(attr))]
        return routes

    @lazy_property
    def type(self):
        if inspect.isfunction(self.real_object):
            return "function"
        elif inspect.isclass(self.real_object):
            return "class"
        elif inspect.ismethod(self.real_object):
            return "method"
        else:
            return "instance" if isinstance(self.real_object, object) else "unknown"

    @lazy_property
    def route_methods(self):
        return {route: getattr(self.real_object, route) for route in self.routes}

    def getattr(self, item):
        if item in self._routes_methods:
            return self._routes_methods[item]
        else:
            raise AttributeError(f"Attribute {item} not served with {self.__class__.__name__}")


class BeamSchema(object):

    def __init__(self, read_schema=None, write_schema=None, **kwargs):
        self._schema = {}
        self._schema.update(kwargs)
        self._read_schema = read_schema or {}
        self._write_schema = write_schema or {}

    def __getitem__(self, item):

        if item in self._schema.keys():
            return self._schema[item]
        elif item in self._read_schema.keys():
            return self._read_schema[item]
        elif item in self._write_schema.keys():
            return self._write_schema[item]
        else:
            raise KeyError(f'No schema for {item}')

    @property
    def read_schema(self):
        return {**self._schema, **self._read_schema}

    @property
    def write_schema(self):
        return {**self._schema, **self._write_schema}



import socket

from ..distributed import ThreadedDispatcher
from ..distributed import RayDispatcher, RayCluster
from .core import BeamFederated
from ..utils import find_port, GPUManager
from ..logger import beam_logger as logger


def federated_executor(func=None, world_size=1, framework='ddp', distributed_backend='nccl', host=None,
                       port=None, func_args=None, func_kwargs=None, kv_store='tcp', kv_store_path=None,
                       kv_store_timeout=300, kv_store_port=None, ray_address=None, ray_kwargs=None, num_gpus=1,
                       num_cpus=4, remote_kwargs=None, **kwargs):

    # ray_cluster = RayCluster(address=ray_address, ray_kwargs=ray_kwargs)
    # if host is None:
    #     host = ray_cluster.head_node_ip
    #     logger.info(f'Host is not specified, using the head node IP: {host}')

    if port is None:
        port = find_port(application='distributed')

    if kv_store_port is None and kv_store == 'tcp':
        kv_store_port = find_port(application='distributed', blacklist=[port])

    logger.info(f'Multiprocessing port is: {port}, KV store port is: {kv_store_port}')

    if kv_store_path is None and kv_store == 'file':
        kv_store_path = '/tmp/beam_kv_store'

    if kv_store_timeout is None:
        kv_store_timeout = 300

    remote_kwargs = remote_kwargs if remote_kwargs is not None else {}
    if num_cpus is not None:
        remote_kwargs['num_cpus'] = num_cpus
    if num_gpus is not None:
        remote_kwargs['num_gpus'] = num_gpus

    logical_devices = None
    if num_gpus is not None:

        hostname = socket.gethostname()

        RayGPUAllocator = RayDispatcher(GPUManager, remote_kwargs={'num_gpus': num_gpus,
                                                             'resources': {f"hostname_{hostname}": 1}},
                                              ray_kwargs=ray_kwargs, asynchronous=False)
        gpu_allocator = RayGPUAllocator()
        logical_devices = GPUManager.logical_devices(physical_devices=gpu_allocator.physical_devices())

    LocalWorkeClass = ThreadedDispatcher(BeamFederated, asynchronous=True)

    remote_workers = [LocalWorkeClass(func=func, rank=0, world_size=world_size, framework=framework,
                                      distributed_backend=distributed_backend, host=host, port=port,
                                      func_args=func_args, func_kwargs=func_kwargs,
                                      kv_store=kv_store, kv_store_path=kv_store_path, devices=logical_devices,
                                      kv_store_timeout=kv_store_timeout, kv_store_port=kv_store_port, **kwargs)]

    if world_size > 1:
        RemoteWorkerClass = RayDispatcher(BeamFederated, address=ray_address, ray_kwargs=ray_kwargs,
                                          remote_kwargs=remote_kwargs, asynchronous=True)

        for rank in range(1, world_size):
            remote_workers.append(RemoteWorkerClass(func=func, rank=rank, world_size=world_size, framework=framework,
                                                    distributed_backend=distributed_backend, host=host, port=port,
                                                    func_args=func_args, func_kwargs=func_kwargs,
                                                    kv_store=kv_store, kv_store_path=kv_store_path,
                                                    kv_store_timeout=kv_store_timeout, kv_store_port=kv_store_port,
                                                    **kwargs))

    return remote_workers







from contextlib import contextmanager
from datetime import datetime
from pathlib import PurePath, Path
from .core import PureBeamPath, normalize_host
from io import StringIO, BytesIO, TextIOWrapper
import os
import urllib3
import stat

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
import pandas as pd
import warnings
from uuid import uuid4 as uuid


class BeamPath(PureBeamPath):

    def __init__(self, *pathsegments, **kwargs):
        scheme = 'file'
        if 'scheme' in kwargs and kwargs['scheme'] is not None:
            scheme = kwargs.pop('scheme')

        PureBeamPath.__init__(self, *pathsegments, scheme=scheme, **kwargs)
        self.path = Path(self.path)

    def glob(self, pattern, case_sensitive=None):
        for path in self.path.glob(pattern):
            yield self.gen(path)

    @classmethod
    def cwd(cls):
        return cls(str(Path.cwd()))

    @classmethod
    def home(cls):
        return cls(str(Path.home()))

    def stat(self):  # add follow_symlinks=False for python 3.10
        return self.path.stat()

    def getmtime(self):
        return os.path.getmtime(str(self.path))

    def getctime(self):
        return os.path.getctime(str(self.path))

    def getatime(self):
        return os.path.getatime(str(self.path))

    def chmod(self, mode):
        return self.path.chmod(mode)

    def exists(self):
        return self.path.exists()

    def expanduser(self):
        return self.path.expanduser()

    def group(self):
        return self.path.group()

    def is_dir(self):
        return self.path.is_dir()

    def is_file(self):
        return self.path.is_file()

    def is_mount(self):
        return self.path.is_mount()

    def is_symlink(self):
        return self.path.is_symlink()

    def is_socket(self):
        return self.path.is_socket()

    def is_fifo(self):
        return self.path.is_fifo()

    def is_block_device(self):
        return self.path.is_block_device()

    def is_char_device(self):
        return self.path.is_char_device()

    def iterdir(self):
        for path in self.path.iterdir():
            yield BeamPath(path)

    def lchmod(self, mode):
        return self.path.lchmod(mode)

    def lstat(self):
        return self.path.lstat()

    def mkdir(self, parents=True, exist_ok=True):
        return self.path.mkdir(parents=parents, exist_ok=exist_ok)

    def owner(self):
        return self.path.owner()

    def read_bytes(self):
        return self.path.read_bytes()

    def read_text(self, *args, **kwargs):
        return self.path.read_text(*args, **kwargs)

    def readlink(self):
        return self.path.readlink()

    def rename(self, target):
        path = self.path.rename(str(target))
        return BeamPath(path)

    def replace(self, target):
        path = self.path.replace(str(target))
        return BeamPath(path)

    def resolve(self, strict=False):

        path = self.path.resolve(strict=strict)
        return self.gen(path)

    def rmdir(self):
        self.path.rmdir()

    def samefile(self, other):
        return self.path.samefile(other)

    def symlink_to(self, target, target_is_directory=False):
        self.path.symlink_to(str(target), target_is_directory=target_is_directory)

    def hardlink_to(self, target):
        self.path.hardlink_to(str(target))

    def link_to(self, target):
        self.path.link_to(str(target))

    def touch(self, *args, **kwargs):
        self.path.touch(*args, **kwargs)

    def unlink(self, missing_ok=False):
        self.path.unlink(missing_ok=missing_ok)

    def write_bytes(self, data):
        return self.path.write_bytes(data)

    def write_text(self, data, *args, **kwargs):
        return self.path.write_text(data, *args, **kwargs)

    def __enter__(self):
        self.file_object = open(self.path, self.mode)
        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_at_exit()

    def write(self, *args, ext=None, **kwargs):

        if ext is None:
            ext = self.suffix

        if ext == '.parquet' and kwargs.get('partition_cols', None) is not None:

            assert len(args) == 1, "Only one argument is allowed for parquet writing with partition_cols"
            df = args[0]
            df.to_parquet(str(self), **kwargs)

        return super().write(*args, ext=ext, **kwargs)

    def read(self, ext=None, **kwargs):

        if ext is None:
            ext = self.suffix

        if ext == '.parquet' and self.is_dir():
            return pd.read_parquet(str(self), **kwargs)

        return super().read(ext=ext, **kwargs)


'''
from smbclient._os import (
    XATTR_CREATE,
    XATTR_REPLACE,
    SMBDirEntry,
    SMBStatResult,
    SMBStatVolumeResult,
    copyfile,
    getxattr,
    link,
    listdir,
    listxattr,
    lstat,
    makedirs,
    mkdir,
    open_file,
    readlink,
    remove,
    removedirs,
    removexattr,
    rename,
    renames,
    replace,
    rmdir,
    scandir,
    setxattr,
    stat,
    stat_volume,
    symlink,
    truncate,
    unlink,
    utime,
    walk,
)

'''


class SMBPath(PureBeamPath):

    def __init__(self, *pathsegments, client=None, hostname=None, username=None, password=None, port=None,
                 connection_timeout=60, **kwargs):

        if port is None:
            port = 445
        super().__init__(*pathsegments, scheme='smb', client=client, hostname=hostname, username=username,
                         password=password, port=port, **kwargs)
        self.connection_timeout = connection_timeout
        if self.client is None:
            import smbclient
            self.client = smbclient
        self.credentials = {'username': self.username, 'password': self.password, 'port': self.port}
        self.register_smb_session()

    @staticmethod
    def _smb_path(server, path):
        path = path.replace('/', '\\')
        return fr"\\{server.upper()}\{path}"

    @property
    def smb_path(self):
        return SMBPath._smb_path(self.hostname, str(self.path))

    def register_smb_session(self):
        self.client.register_session(self.hostname.upper(), connection_timeout=self.connection_timeout,
                                     **self.credentials)

    def exists(self):
        try:
            self.client.stat(self.smb_path)
            return True
        except FileNotFoundError:
            # The path does not exist
            return False
        except Exception as e:
            # Other exceptions might indicate a problem with the network, permissions, etc.
            # Depending on your application, you might want to handle these differently.
            return False

    def is_file(self):
        try:
            file_attributes = self.client.stat(self.smb_path).st_file_attributes
            # Check if the directory attribute is not set
            return not file_attributes & stat.FILE_ATTRIBUTE_DIRECTORY
        except Exception as e:
            # Handle exceptions, possibly returning False or re-raising
            return False

    def is_dir(self):
        try:
            file_attributes = self.client.stat(self.smb_path).st_file_attributes
            # Check if the directory attribute is set
            return bool(file_attributes & stat.FILE_ATTRIBUTE_DIRECTORY)
        except Exception as e:
            # Handle exceptions
            return False

    def mkdir(self, parents=True, exist_ok=True, **kwargs):

        if self.is_root():
            return

        if not exist_ok:
            if self.exists():
                raise FileExistsError(f"File already exists: {self.smb_path}")

        if parents and not self.parent.exists():
            self.parent.mkdir(parents=True, exist_ok=True, **kwargs)

        from smbprotocol.exceptions import SMBOSError
        from smbprotocol.header import NtStatus

        try:
            self.client.mkdir(self.smb_path, **{**self.credentials, **kwargs})
        except SMBOSError as e:
            if e.ntstatus == NtStatus.STATUS_OBJECT_NAME_COLLISION:
                pass
            else:
                raise e

    def rmdir(self):
        self.client.rmdir(self.smb_path, **self.credentials)

    def unlink(self, missing_ok=False):
        self.client.unlink(self.smb_path, **self.credentials)

    def iterdir(self):
        for p in self.client.listdir(self.smb_path, **self.credentials):
            yield self.gen(f"{self.path}/{p}")

    def replace(self, target):
        self.rename(target)

    def rename(self, target):
        self.client.rename(self.smb_path, target.smb_path, **self.credentials)

    def __enter__(self):
        if self.mode in ["rb", "r"]:
            self.file_object = self.client.open_file(self.smb_path, mode=self.mode, **self.credentials,
                                                     encoding=self.open_kwargs['encoding'],
                                                     newline=self.open_kwargs['newline'],
                                                     errors=self.open_kwargs['errors'])
        elif self.mode == 'wb':
            self.file_object = BytesIO()
        elif self.mode == 'w':
            self.file_object = StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError
        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.mode in ["wb", "w"]:
            self.file_object.seek(0)
            self.client.open_file(self.smb_path, mode=self.mode, **self.credentials).write(self.file_object.getvalue())
        self.close_at_exit()


class SFTPPath(PureBeamPath):

    def __init__(self, *pathsegments, client=None, hostname=None, username=None, private_key=None, password=None,
                 port=None, private_key_pass=None, ciphers=None, log=False, cnopts=None, default_path=None,
                 disable_hostkey=True, **kwargs):

        super().__init__(*pathsegments, scheme='sftp', client=client, hostname=hostname, username=username,
                         private_key=private_key, password=password, port=port, private_key_pass=private_key_pass,
                         ciphers=ciphers, log=log, cnopts=cnopts, default_path=default_path, **kwargs)

        if port is None:
            port = 22
        elif isinstance(port, str):
            port = int(port)

        if client is None:
            import pysftp
            if disable_hostkey:
                cnopts = pysftp.CnOpts()
                cnopts.hostkeys = None  # This disables host key checking
            self.client = pysftp.Connection(host=hostname, username=username, private_key=private_key, password=password,
                                            port=port, private_key_pass=private_key_pass, ciphers=ciphers, log=log,
                                            cnopts=cnopts, default_path=default_path)
        else:
            self.client = client

    def samefile(self, other):
        raise NotImplementedError

    def iterdir(self):

        for p in self.client.listdir(remotepath=str(self.path)):
            path = self.path.joinpath(p)
            yield self.gen(path)

    def is_file(self):
        return self.client.isfile(remotepath=str(self.path))

    def is_dir(self):
        return self.client.isdir(remotepath=str(self.path))

    def mkdir(self, *args, mode=777, **kwargs):
        self.client.makedirs(str(self.path), mode=mode)

    def exists(self):
        return self.client.exists(str(self.path))

    def rename(self, target):
        self.client.rename(str(self.path), str(target))

    def __enter__(self):
        if self.mode == "rb":
            self.file_object = self.client.open(str(self.path), self.mode)
        elif self.mode == "r":
            self.file_object = TextIOWrapper(self.client.open(str(self.path), self.mode),
                                             encoding=self.open_kwargs['encoding'],
                                            newline=self.open_kwargs['newline'])
        elif self.mode == 'wb':
            self.file_object = BytesIO()
        elif self.mode == 'w':
            self.file_object = StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError
        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:
            self.file_object.seek(0)
            self.client.putfo(self.file_object, remotepath=str(self.path))

        self.close_at_exit()

    def rmdir(self):
        self.client.rmdir(str(self.path))

    def unlink(self, missing_ok=False):

        if self.is_file():
            self.client.remove(str(self.path))
        else:
            raise FileNotFoundError


class S3Path(PureBeamPath):

    def __init__(self, *pathsegments, client=None, hostname=None, port=None, access_key=None,
                 secret_key=None, tls=True, storage_class=None, **kwargs):
        super().__init__(*pathsegments, scheme='s3', client=client, hostname=hostname, port=port,
                         access_key=access_key, secret_key=secret_key, tls=tls, storage_class=storage_class, **kwargs)

        if not self.is_absolute():
            self.path = PurePath('/').joinpath(self.path)

        if len(self.parts) > 1:
            self.bucket_name = self.parts[1]
        else:
            self.bucket_name = None

        if len(self.parts) > 2:
            self.key = '/'.join(self.parts[2:])
        else:
            self.key = None

        if client is None:

            import boto3

            if type(tls) is str:
                tls = tls.lower() == 'true'

            protocol = 'https' if tls else 'http'
            kwargs = {}
            if hostname is not None:
                kwargs['endpoint_url'] = f'{protocol}://{normalize_host(hostname, port)}'

            if hostname is None and 'region_name' not in kwargs:
                warnings.warn("When working with AWS, please define region_name in kwargs to avoid extra cost")

            client = boto3.resource(config=boto3.session.Config(signature_version='s3v4'),
                                    verify=False, service_name='s3', aws_access_key_id=access_key,
                                    aws_secret_access_key=secret_key, **kwargs)

        self.client = client
        self._bucket = None
        self._object = None
        self.storage_class = storage_class or 'STANDARD'

    @property
    def bucket(self):

        if self.bucket_name is None:
            self._bucket = None
        elif self._bucket is None:
            self._bucket = self.client.Bucket(self.bucket_name)
        return self._bucket

    @property
    def object(self):
        if self._object is None:
            self._object = self.client.Object(self.bucket_name, self.key)
        return self._object

    def is_file(self):

        if self.bucket_name is None or self.key is None:
            return False

        key = self.key.rstrip('/')
        return S3Path._exists(self.client, self.bucket_name, key)

    @staticmethod
    def _exists(client, bucket_name, key):
        import botocore
        try:
            # client.Object(bucket_name, key).load()
            client.meta.client.head_object(Bucket=bucket_name, Key=key)
            return True
        except botocore.exceptions.ClientError:
            return False

    def is_dir(self):

        if self.bucket_name is None:
            return True

        if self.key is None:
            return self._check_if_bucket_exists()

        key = self.normalize_directory_key()
        return S3Path._exists(self.client, self.bucket_name, key) or \
               (self._check_if_bucket_exists() and (not self._is_empty(key)))

    def read_text(self, encoding=None, errors=None):
        return self.object.get()["Body"].read().decode(encoding, errors)

    def read_bytes(self):
        return self.object.get()["Body"].read()

    def exists(self):

        if self.key is None:
            return self._check_if_bucket_exists()
        return S3Path._exists(self.client, self.bucket_name, self.key) or self.is_dir()

    def rename(self, target):
        self.object.copy_from(
            CopySource={
                "Bucket": self.bucket_name,
                "Key": self.key,
            },
            Bucket=target.bucket_name,
            Key=target.key,
        )
        self.unlink()

    def _check_if_bucket_exists(self):
        try:
            self.client.meta.client.head_bucket(Bucket=self.bucket_name)
            return True
        except self.client.meta.client.exceptions.ClientError:
            return False

    def replace(self, target):
        self.rename(target)

    def unlink(self, **kwargs):
        if self.is_file():
            self.object.delete()
        if self.is_dir():
            obj = self.client.Object(self.bucket_name, f"{self.key}/")
            obj.delete()

    def mkdir(self, parents=True, exist_ok=True):

        if not parents:
            raise NotImplementedError("parents=False is not supported")

        if exist_ok and self.exists():
            return

        if not self._check_if_bucket_exists():
            self.bucket.create()

        if self.key is not None:
            key = self.normalize_directory_key()
            self.bucket.put_object(Key=key, StorageClass=self.storage_class)

    def _is_empty_bucket(self):
        for _ in self.bucket.objects.all():
            return False
        return True

    def _is_empty(self, key=None):
        if key is None:
            key = self.key
        for obj in self.bucket.objects.filter(Prefix=key):
            if obj.key.rstrip('/') != self.key.rstrip('/'):
                return False
        return True

    def rmdir(self):

        if self.key is None:
            if not self._is_empty_bucket():
                raise OSError("Directory not empty: %s" % self)
            self.bucket.delete()

        else:
            if self.is_file():
                raise NotADirectoryError("Not a directory: %s" % self)

            if not self._is_empty():
                raise OSError("Directory not empty: %s" % self)

            self.unlink()
            # self.bucket.delete_objects(Delete={"Objects": [{"Key": path.key} for path in self.iterdir()]})

    def key_depth(self, key=None):
        if key is None:
            key = self.key
        if key is None:
            return 0
        return len(list(filter(lambda x: len(x), key.split('/'))))

    def normalize_directory_key(self, key=None):
        if key is None:
            key = self.key
        if key is None:
            return None
        if not key.endswith('/'):
            key += '/'
        return key

    def iterdir(self):

        if self.bucket is None:
            for bucket in self.client.buckets.all():
                yield self.gen(bucket.name)
            return

        key = self.normalize_directory_key()
        if key is None:
            key = ''

        # objects = self.client.meta.client.list_objects_v2(Bucket=self.bucket_name, Prefix=key, Delimiter='/')

        paginator = self.client.meta.client.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(Bucket=self.bucket_name, Prefix=key, Delimiter='/')

        for objects in page_iterator:
            if 'CommonPrefixes' in objects:
                for prefix in objects['CommonPrefixes']:
                    path = f"{self.bucket_name}/{prefix['Prefix']}"
                    yield self.gen(path)

            if 'Contents' in objects:
                for content in objects['Contents']:
                    if content['Key'] == key:
                        continue
                    path = f"{self.bucket_name}/{content['Key']}"
                    yield self.gen(path)

    def __enter__(self):
        if self.mode in ["rb", "r"]:
            file_object = self.client.meta.client.get_object(Bucket=self.bucket_name, Key=self.key)['Body']
            # io_obj = StringIO if 'r' else BytesIO
            self.file_object = BytesIO(file_object.read())
        elif self.mode == 'wb':
            self.file_object = BytesIO()
        elif self.mode == 'w':
            self.file_object = StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError

        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:
            self.file_object.seek(0)
            self.client.Object(self.bucket_name, self.key).put(Body=self.file_object.getvalue(),
                                                               StorageClass=self.storage_class)

        self.close_at_exit()

    def getmtime(self):
        d = self.object.get()["LastModified"]
        return d.timestamp()


class PyArrowPath(PureBeamPath):

    def __init__(self, *args, strip_path=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.strip_path = strip_path

    @property
    def str_path(self):
        if self.strip_path:
            return str(self).lstrip('/')
        return str(self)

    @property
    def file_info(self):
        return self.client.get_file_info([self.str_path])[0]

    def _exists(self, dir=False, file=False):
        from pyarrow.lib import ArrowIOError
        from pyarrow import fs
        try:
            fi = self.file_info
            if dir:
                return fi.type == fs.FileType.Directory
            if file:
                return fi.type == fs.FileType.File
            return True
        except ArrowIOError:
            return False

    def is_file(self):
        return self._exists(file=True)

    def is_dir(self):
        return self._exists(dir=True)

    def exists(self):
        return self._exists()

    def rename(self, target):
        self.client.move(self.str_path, self.str_path)

    def replace(self, target):
        self.rename(target)

    def unlink(self, **kwargs):
        self.client.delete_file(self.str_path)

    def rmtree(self):
        if self.is_file():
            self.unlink()
        else:
            self.client.delete_dir_contents(self.str_path)

    def mkdir(self, parents=True, exist_ok=True):
        if parents:
            self.client.create_dir(self.str_path, recursive=True)
        else:
            self.client.create_dir(self.str_path, recursive=False)

    def rmdir(self):
        self.client.delete_dir(self.str_path)

    def iterdir(self):

        from pyarrow import fs
        fi = self.client.get_file_info(fs.FileSelector(self.str_path, recursive=False))
        for f in fi:
            yield self.gen(f.path)

    def __enter__(self):
        if self.mode in ["rb", "r"]:
            with self.client.open_input_file(self.str_path) as f:
                content = f.read()
            # io_obj = StringIO if 'r' else BytesIO
            encoding = self.open_kwargs['encoding'] or 'utf-8'
            self.file_object = BytesIO(content) if 'b' in self.mode else StringIO(content.decode(encoding),
                                                                                  newline=self.open_kwargs['newline'])
        elif self.mode in ['wb', 'w']:
            self.file_object = BytesIO() if 'b' in self.mode else StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError("Invalid mode")

        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:
            self.file_object.seek(0)
            content = self.file_object.getvalue()
            with self.client.open_output_stream(self.str_path) as f:
                f.write(content if 'b' in self.mode else content.encode())

        self.close_at_exit()

    def write(self, *args, ext=None, **kwargs):

        x = None
        if len(args) >= 1:
            x = args[0]

        if ext is None:
            ext = self.suffix

        if ext == '.parquet':
            import pyarrow.parquet as pq
            pq.write_table(x, self.str_path, filesystem=self.client)

        elif ext == '.orc':
            import pyarrow.orc as orc
            orc.write_table(x, self.str_path, filesystem=self.client)

        return super().write(*args, ext=ext, **kwargs)

    def read(self, ext=None, **kwargs):

        if ext is None:
            ext = self.suffix

        if ext == '.parquet':
            import pyarrow.parquet as pq
            return pq.read_table(self.str_path, filesystem=self.client)

        if ext == '.orc':
            import pyarrow.orc as orc
            return orc.read_table(self.str_path, filesystem=self.client)

        return super().read(ext=ext, **kwargs)


class HDFSPath(PureBeamPath):

    def __init__(self, *pathsegments, client=None, hostname=None, port=None, timeout=None,
                 username=None, skip_trash=False, n_threads=0,  temp_dir=None, chunk_size=65536,
                 progress=None, cleanup=True, tls=True, **kwargs):
        super().__init__(*pathsegments, scheme='hdfs', hostname=hostname, port=port, skip_trash=skip_trash,
                                        username=username, n_threads=n_threads, temp_dir=temp_dir, timeout=timeout,
                                        chunk_size=chunk_size, progress=progress, cleanup=cleanup, **kwargs)

        from hdfs import InsecureClient

        if type(tls) is str:
            tls = tls.lower() == 'true'

        protocol = 'https' if tls else 'http'

        if client is None:
            client = InsecureClient(f'{protocol}://{normalize_host(hostname, port)}', user=username)

        self.client = client

    def exists(self):
        return self.client.status(str(self), strict=False) is not None

    def rename(self, target):
        self.client.rename(str(self), str(target))

    def replace(self, target):

        self.client.rename(str(self), str(target))
        return HDFSPath(target, client=self.client)

    def unlink(self, missing_ok=False):
        if not missing_ok:
            self.client.delete(str(self), skip_trash=self['skip_trash'])
        self.client.delete(str(self), skip_trash=self['skip_trash'])

    def mkdir(self, mode=0o777, parents=True, exist_ok=True):
        if not exist_ok:
            if self.exists():
                raise FileExistsError
        if not parents:
            raise NotImplementedError('parents=False not implemented for HDFSPath.mkdir')
        self.client.makedirs(str(self), permission=mode)

    def rmdir(self):
        self.client.delete(str(self), skip_trash=self['skip_trash'])

    def iterdir(self):
        files = self.client.list(str(self))
        for f in files:
            yield self.joinpath(f)

    def samefile(self, other):
        raise NotImplementedError

    def is_file(self):

        status = self.client.status(str(self), strict=False)
        if status is None:
            return False
        return status['type'] == 'FILE'

    def is_dir(self):

        status = self.client.status(str(self), strict=False)
        if status is None:
            return False
        return status['type'] == 'DIRECTORY'

    def read(self, ext=None, **kwargs):

        if ext is None:
            ext = self.suffix

        if ext == '.avro':
            from hdfs.ext.avro import AvroReader
            x = []
            with AvroReader(self.client, str(self), **kwargs) as reader:
                # reader.writer_schema  # The remote file's Avro schema.
                # reader.content  # Content metadata (e.g. size).
                for record in reader:
                    x.append(record)
            return x

        elif ext == '.pd':
            from hdfs.ext.dataframe import read_dataframe
            return read_dataframe(self.client, str(self))

        return super().read(ext=ext, **kwargs)

    def write(self, *args, ext=None,  **kwargs):

        x = None
        if len(args) >= 1:
            x = args[0]

        if ext is None:
            ext = self.suffix

        if ext == '.avro':
            from hdfs.ext.avro import AvroWriter
            with AvroWriter(self.client, str(self)) as writer:
                for record in x:
                    writer.write(record)

        elif ext == '.pd':
            from hdfs.ext.dataframe import write_dataframe
            write_dataframe(self.client, str(self), x, **kwargs)

        else:
            super().write(*args, ext=ext, **kwargs)

    def __enter__(self):
        if self.mode in ["rb", "r"]:

            # chunk_size = self.query['chunk_size']
            # chunk_size = int(chunk_size) if chunk_size is not None else None
            # content = self.client.read(str(self), chunk_size=chunk_size)

            chunk_size = self.query.get('chunk_size', None)
            chunk_size = int(chunk_size) if chunk_size is not None else 0

            with self.client.read(str(self), chunk_size=chunk_size) as reader:
                content = reader.read()

            encoding = self.open_kwargs['encoding'] or 'utf-8'
            self.file_object = BytesIO(content) if 'b' in self.mode else StringIO(content.decode(encoding),
                                                                                  newline=self.open_kwargs['newline'])

        elif self.mode in ['wb', 'w']:
            self.file_object = BytesIO() if 'b' in self.mode else StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError

        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:

            self.file_object.seek(0)
            content = self.file_object.getvalue()
            with self.client.write(str(self)) as writer:
                writer.write(content if 'b' in self.mode else content.encode())

        self.close_at_exit()


class S3PAPath(PyArrowPath):
    # a pyarrow implementation of S3Path
    def __init__(self, *pathsegments, client=None, hostname=None, port=None, access_key=None,
                 secret_key=None, tls=True, **kwargs):
        super().__init__(*pathsegments, scheme='s3-pa', client=client, hostname=hostname, port=port,
                         access_key=access_key, secret_key=secret_key, tls=tls, strip_path=True, **kwargs)

        if client is None:

            from pyarrow import fs

            if hostname is not None:
                kwargs['endpoint_override'] = normalize_host(hostname, port)

            if hostname is None and 'region' not in kwargs:
                warnings.warn("When working with AWS, please define region_name in kwargs to avoid extra cost")

            if type(tls) is str:
                tls = tls.lower() == 'true'

            kwargs['scheme'] = 'https' if tls else 'http'

            if 'allow_bucket_creation' not in kwargs:
                kwargs['allow_bucket_creation'] = True
            if 'allow_bucket_deletion' not in kwargs:
                kwargs['allow_bucket_deletion'] = True
            # kwargs['use_virtual_addressing'] = False

            client = fs.S3FileSystem(access_key=access_key, secret_key=secret_key, **kwargs)

        self.client = client


class HDFSPAPath(PyArrowPath):

    # a pyarrow implementation of HDFSPath
    def __init__(self, *pathsegments, client=None, hostname=None, port=None,  username=None, buffer_size=0,
                 replication=3, kerb_ticket=None, extra_conf=None, default_block_size=None, **kwargs):

        super(HDFSPAPath).__init__(*pathsegments, scheme='hdfs-pa', hostname=hostname, port=port,
                         username=username, **kwargs)

        if client is None:
            from pyarrow import fs
            client = fs.HadoopFileSystem(hostname, port=int(port), user=username, replication=replication,
                                         buffer_size=buffer_size, default_block_size=default_block_size,
                                         kerb_ticket=kerb_ticket, extra_conf=extra_conf)

        self.client = client


@contextmanager
def temp_local_file(content, tmp_path='/tmp', name=None, ext=None, binary=True, as_beam_path=True):
    tmp_path = BeamPath(tmp_path).joinpath(uuid())
    tmp_path.mkdir(exist_ok=True, parents=True)
    if name is None:
        name = uuid()
    if ext is not None:
        name = f"{name}{ext}"
    tmp_path = tmp_path.joinpath(name)
    try:
        if binary:
            tmp_path.write_bytes(content)
        else:
            tmp_path.write_text(content)
        yield tmp_path if as_beam_path else str(tmp_path)
    finally:
        tmp_path.unlink()
        tmp_path.parent.rmdir()


class CometAsset(PureBeamPath):
    # a pathlib/beam_path api for comet artifacts

    def __init__(self, *pathsegments, client=None, hostname=None, port=None, access_key=None,
                 secret_key=None, tls=True, **kwargs):
        super().__init__(*pathsegments, scheme='comet', client=client, hostname=hostname, port=port,
                         access_key=access_key, secret_key=secret_key, tls=tls, **kwargs)

        if self.hostname is not None:
            tls = 'https' if tls else 'http'
            os.environ['COMET_URL_OVERRIDE'] = f"{tls}://{normalize_host(self.hostname, self.port)}/clientlib"

        if client is None:

            from comet_ml import API
            client = API(api_key=access_key)

        self.client = client
        parts = self.parts[1:]

        self.workspace = None
        self.project_name = None
        self.experiment_name = None
        self.asset_name = None
        self._experiment_key = None
        self.level = len(parts)
        if self.level > 0:
            self.workspace = parts[0]
        if self.level > 1:
            self.project_name = parts[1]
        if self.level > 2:
            self.experiment_name = parts[2]
        if self.level > 3:
            self.asset_name = parts[3]

    @property
    def experiment_key(self):
        if self.level < 3:
            return None
        if self._experiment_key is None:
            experiments = self.client.get_experiments(self.workspace, self.project_name)
            for e in experiments:
                if e.name == self.experiment_name:
                    self._experiment_key = e.key
                    break
        return self._experiment_key


    @property
    def assets_map(self):
        if self.level < 2:
            return None
        assets_map = self.experiment.get_asset_list()
        return {asset['fileName']: asset for asset in assets_map}

    @property
    def next_level(self):
        kwargs = {}
        args = ()
        attr = 'get'
        if self.level > 0:
            attr = 'get'
            kwargs['workspace'] = self.workspace
        if self.level > 1:
            attr = 'get_experiments'
            kwargs['project_name'] = self.project_name
        if self.level > 2:
            attr = 'get_experiment'
            kwargs = {}
            args = (self.workspace, self.project_name, self.experiment_key,)
        if self.level > 3:
            raise ValueError("CometArtifact: too many levels, it is not a directory")
        return getattr(self.client, attr)(*args, **kwargs)

    def is_file(self):
        assets_map = self.assets_map
        if assets_map is None or self.name not in assets_map:
            return False
        return True

    def is_dir(self):
        return self.level < 4 and self.assets_map is not None

    def exists(self):
        assets_map = self.assets_map
        if assets_map is None:
            return False
        if self.level < 4:
            return True
        return self.name in assets_map

    def iterdir(self):
        if self.level in [0, 1]:
            for p in self.next_level:
                path = self.path.joinpath(p)
                yield self.gen(path)
        elif self.level == 2:
            for e in self.next_level:
                path = self.path.joinpath(e.name)
                yield self.gen(path)
        elif self.level == 3:
            assets_map = self.assets_map
            if assets_map is None:
                return
            for a in assets_map:
                path = self.path.joinpath(a)
                yield self.gen(path)
        else:
            raise ValueError("CometArtifact: too many levels, it is not a directory")

    @property
    def experiment(self):
        if self.level < 3:
            return None
        return self.client.get_experiment(self.workspace, self.project_name, self.experiment_key)

    def mkdir(self, *args, **kwargs):

        if self.level == 1:
            raise ValueError("CometArtifact: cannot create workspace")
        elif self.level == 2:
            self.client.create_project(self.workspace, self.project_name)
        elif self.level == 3:
            from comet_ml import APIExperiment
            exp = APIExperiment(api_key=self.client.api_key, workspace=self.workspace, project_name=self.project_name)
            exp.set_name(self.experiment_name)

        ValueError("CometArtifact: cannot create a directory at this hierarchy level")

    def rmdir(self):
        raise NotImplementedError

    def unlink(self, missing_ok=False):
        if missing_ok:
            raise NotImplementedError
        if self.level == 1:
            raise ValueError("CometArtifact: cannot delete workspace")
        elif self.level == 2:
            self.client.delete_project(self.project_name)
        elif self.level == 3:
            self.client.delete_experiment(self.next_level.key)
        elif self.level == 4:
            self.next_level.delete_asset(self.assets_map[self.name]['assetId'])
        else:
            raise ValueError("CometArtifact: cannot delete an object at this hierarchy level")

    def rename(self, target):
        raise NotImplementedError

    def replace(self, target):
        raise NotImplementedError

    def __enter__(self):
        if self.mode in ["rb", "r"]:

            content = self.experiment.get_asset(self.assets_map[self.name]['assetId'])
            encoding = self.open_kwargs['encoding'] or 'utf-8'
            self.file_object = BytesIO(content) if 'b' in self.mode else StringIO(content.decode(encoding),
                                                                                  newline=self.open_kwargs['newline'])

        elif self.mode in ['wb', 'w']:
            self.file_object = BytesIO() if 'b' in self.mode else StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError

        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:
            self.file_object.seek(0)
            content = self.file_object.getvalue()
            with temp_local_file(content, name=self.name, as_beam_path=True, binary='b' in self.mode) as tmp_path:
                cwd = os.getcwd()
                try:
                    os.chdir(tmp_path.parent)
                    self.experiment.log_asset(tmp_path.name)
                finally:
                    os.chdir(cwd)
        self.close_at_exit()


class RedisPath(PureBeamPath):

    def __init__(self, *pathsegments, client=None, hostname=None, port=None, password=None, username=None, data=None,
                 tls=False, db=0, **kwargs):
        super().__init__(*pathsegments, scheme='redis', client=client, data=data, **kwargs)

        if type(tls) is str:
            tls = tls.lower() == 'true'

        if client is None:

            if port is None:
                port = 6379
            if hostname is None:
                hostname = 'localhost'

            import redis
            client = redis.Redis(host=hostname, port=port, db=db, password=password, username=username, ssl=tls)

        self.client = client

    def normalize_directory_key(self, key=None):
        if key is None:
            key = self.key
        if key is None:
            return None
        if not key.endswith('/'):
            key = f'{key}/'
        return key

    def mkdir(self, *args, parents=True, exist_ok=True):

        if self.is_root():
            return

        if not exist_ok:
            if self.exists():
                raise FileExistsError

        if not parents:
            if not self.parent.exists():
                raise FileNotFoundError
        else:
            p = self.parent
            while not p.root:
                if not p.exists():
                    p.mkdir()
                else:
                    break
                p = p.parent

        self.write_to_redis(self.client, self.directory_key, '')

    def rmdir(self):
        if not self.is_dir():
            raise NotADirectoryError
        if not self.is_empty():
            raise OSError("Directory not empty: %s" % self)
        self.client.delete(self.directory_key)

    def unlink(self, missing_ok=False):
        if self.is_dir():
            raise IsADirectoryError
        self.client.delete(self.key)

    def is_file(self):

        if self.key.endswith('/'):
            return False
        return bool(self.client.exists(self.key))

    def is_dir(self):
        return bool(self.client.exists(self.directory_key))

    def exists(self):
        return bool(self.client.exists(self.key) or self.client.exists(self.directory_key))

    def iterdir(self):
        if not self.is_dir():
            raise NotADirectoryError

        prefix = self.directory_key
        for key in self.client.scan_iter(f'{prefix}*'):
            key = key.decode('utf-8')
            if key.count('/') == prefix.count('/') or (key.endswith('/') and key.count('/') == prefix.count('/') + 1):

                if prefix.rstrip('/') == key.rstrip('/'):
                    continue
                yield self.gen(key)

        # # Pattern for files
        # file_pattern = f'{prefix}[^/]+'
        # for key in self.client.scan_iter(file_pattern):
        #     yield self.gen(key.decode('utf-8'))
        #
        # # Pattern for directories
        # dir_pattern = f'{prefix}[^/]+/'
        # for key in self.client.scan_iter(dir_pattern):
        #     yield self.gen(key.decode('utf-8'))

    def is_empty(self):
        for _ in self.iterdir():
            return False
        return True

    def rename(self, target):
        self.client.rename(self.key, target.key)

    def replace(self, target):
        if target.exists():
            target.unlink()
        self.rename(target)

    @property
    def directory_key(self):
        return self.normalize_directory_key(self.key)

    @property
    def key(self):
        # key = '/'.join(self.parts)
        # # to remove the leading slash if exists
        # key = key.lstrip('/')
        key = str(self)
        return key

    @property
    def _obj(self):
        obj = self.client.hget(self.key, 'data')
        return obj

    @property
    def _timestamp(self):
        timestamp = self.client.hget(self.key, 'modified')
        return timestamp

    def __enter__(self):
        if self.mode in ["rb", "r"]:
            content = self._obj
            encoding = self.open_kwargs['encoding'] or 'utf-8'
            self.file_object = BytesIO(content) if 'b' in self.mode else StringIO(content.decode(encoding),
                                                                                  newline=self.open_kwargs['newline'])
        elif self.mode in ['wb', 'w']:
            self.file_object = BytesIO() if 'b' in self.mode else StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError

        return self.file_object

    @staticmethod
    def write_to_redis(client, key, content, timestamp=None):
        if timestamp is None:
            timestamp = str(datetime.now())
        client.hset(key, 'data', content)
        client.hset(key, 'modified', timestamp)

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ['wb', 'w']:
            self.file_object.seek(0)
            content = self.file_object.getvalue()
            self.write_to_redis(self.client, self.key, content)

        self.close_at_exit()

    def glob(self, pattern, case_sensitive=None):
        full_pattern = f'{self.directory_key}{pattern}'
        return [self.gen(key) for key in self.client.scan_iter(full_pattern)]



from argparse import Namespace

from ...resource import beam_llm


def stream_openai_generator(response):
    for res in response:
        yield res.openai_format


def simulate_openai_chat(model=None, stream=False, **kwargs):
    llm = beam_llm(model) if type(model) == str else model
    res = llm.chat_completion(stream=stream, **kwargs)
    if stream:
        return stream_openai_generator(res)
    else:
        return res.openai_format


def simulate_openai_completion(model=None, **kwargs):
    llm = beam_llm(model) if type(model) == str else model
    return llm.completion(**kwargs).openai_format


class OpenAI:

    chat = Namespace(completions=Namespace(create=simulate_openai_chat))
    completions = Namespace(create=simulate_openai_completion)

    def __init__(self, **kwargs):
       pass


import time
import random
from faker import Faker
from ..core import Processor


class BeamFakeAlg(Processor):
    def __init__(self, sleep_time=1, variance=0., error_rate=0., **faker_fields):
        """
        Initialize the BeamFakeAlg.

        :param sleep_time: Base sleep time in seconds.
        :param variance: Variance in sleep time in seconds.
        :param error_rate: Probability of raising an exception to simulate errors.
        :param faker_fields: Configurable fields for Faker to generate.
        """
        super().__init__()
        self.sleep_time = sleep_time
        self.variance = variance
        self.error_rate = error_rate
        self.faker_fields = faker_fields if len(faker_fields) > 0 else {'name': 'name', 'email': 'email',
                                                                        'address': 'address'}
        self.fake = Faker()

    def run(self, x):
        """
        Process the input data and return fake results.

        :param x: Input data.
        :return: Fake data results or raises an exception.
        """
        # Simulate processing time
        actual_sleep_time = self.sleep_time + random.uniform(-self.variance, self.variance)
        time.sleep(actual_sleep_time)

        # Simulate error
        if random.random() < self.error_rate:
            raise Exception("Simulated error")

        # Generate fake data based on input x (not used in this simple example)
        # You can modify this part to incorporate 'x' in the data generation logic
        data = {field: getattr(self.fake, field)() for field in self.faker_fields}

        return data


from ..config import BeamConfig, BeamParam
import os


class RayConfig(BeamConfig):
    parameters = [
        BeamParam('include-dashboard', bool, True, 'include ray-dashboard', ),
        BeamParam('runtime-env', str, None, 'runtime environment for ray', ),
        BeamParam('dashboard-port', int, None, 'dashboard port for ray', ),
        BeamParam('ray-address', str, 'auto', 'whether to link to existing ray cluster (auto/ip) or to '
                                              'set up a local ray instance (local/ip)', ),
    ]


class HPOConfig(RayConfig, BeamConfig):

    parameters = [
        BeamParam('gpus-per-trial', int, 1, 'number of gpus per trial',),
        BeamParam('cpus-per-trial', int, 4, 'number of cpus per trial',),
        BeamParam('num-trials', int, 1000, 'number of HPO trails',),
        BeamParam('n-jobs', int, 1, 'number of parallel HPO jobs',),
        BeamParam('time-budget-s', int, None, 'time budget in seconds',),
        BeamParam('print-results', bool, False, 'print the intermediate results during training',),
        BeamParam('enable-tqdm', bool, False, 'enable tqdm progress bar',),
        BeamParam('print-hyperparameters', bool, True, 'print the hyperparameters before training',),
        BeamParam('verbose', bool, True, 'verbose mode in hyperparameter optimization',),
        BeamParam('track-results', bool, False, 'track the results of each trial',),
        BeamParam('track-algorithms', bool, False, 'track the algorithms of each trial',),
        BeamParam('track-hparams', bool, True, 'track the hyperparameters of each trial',),
        BeamParam('track-suggestion', bool, True, 'track the suggestions of each trial',),
        BeamParam('hpo-path', str, os.path.join(os.path.expanduser('~'), 'beam_projects', 'hpo'),
                  'Root directory for Logs and results of Hyperparameter optimizations and the associated experiments'),
        BeamParam('stop', str, None, 'stop criteria for the HPO',),
        BeamParam('get-port-from-beam-port-range', bool, True, 'get port from beam port range',),
        BeamParam('replay-buffer-size', int, None, 'Maximal size of finite-memory hpo',),
        BeamParam('time-window', int, None, 'Maximal time window of finite-memory hpo',),

    ]




from collections import defaultdict
from functools import partial

from ..core import Processor
from ..utils import lazy_property, as_numpy, check_type, as_tensor

triton_to_numpy_dtype_dict = {
    'BOOL': 'bool',
    'INT8': 'int8',
    'INT16': 'int16',
    'INT32': 'int32',
    'INT64': 'int64',
    'UINT8': 'uint8',
    'UINT16': 'uint16',
    'UINT32': 'uint32',
    'UINT64': 'uint64',
    'FP16': 'float16',
    'FP32': 'float32',
    'FP64': 'float64',
    'BYTES': 'bytes',
    'STRING': 'string',
    'UNDEFINED': 'undefined',
}


class TritonClient(Processor):

    def __init__(self, scheme='http', host='localhost', port=8000, model_name=None, model_version=None,
                 verbose=False, concurrency=1, connection_timeout=60.0, network_timeout=60.,
                 max_greenlets=None, ssl_options=None, *args, ssl_context_factory=None,
                 insecure=False, config=None, **kwargs):
        super().__init__(*args, **kwargs)

        self.scheme = scheme
        self.host = host
        self.port = port
        self.model_name = model_name
        self.model_version = model_version
        self.verbose = verbose
        self.concurrency = concurrency
        self.connection_timeout = connection_timeout
        self.network_timeout = network_timeout
        self.max_greenlets = max_greenlets
        self.ssl_options = ssl_options
        self.ssl_context_factory = ssl_context_factory
        self.insecure = insecure
        self.ssl = scheme == 'https' or scheme == 'grpcs'
        self.config = config
        self.metadata_cache = defaultdict(dict)

    def get_metadata(self, model_name=None, model_version=None):

        if model_name in self.metadata_cache and model_version in self.metadata_cache[model_name]:
            return self.metadata_cache[model_name][model_version]

        model_name = model_name or self.model_name
        model_version = model_version or self.model_version or ''
        metadata = self.client.get_model_metadata(model_name, model_version=model_version)

        self.metadata_cache[model_name][model_version] = metadata
        return metadata

    @lazy_property
    def metadata(self):
        return self.get_metadata(self.model_name, self.model_version)

    @property
    def is_alive(self):
        return self.client.is_server_live()

    @lazy_property
    def infer_input(self):
        if 'http' in self.scheme:
            from tritonclient.http import InferInput
        elif 'grpc' in self.scheme:
            from tritonclient.grpc import InferInput
        else:
            raise ValueError(f"Invalid scheme: {self.scheme}")
        return InferInput

    @lazy_property
    def infer_requested_output(self):
        if 'http' in self.scheme:
            from tritonclient.http import InferRequestedOutput
        elif 'grpc' in self.scheme:
            from tritonclient.grpc import InferRequestedOutput
        else:
            raise ValueError(f"Invalid scheme: {self.scheme}")
        return InferRequestedOutput

    @lazy_property
    def client(self):
        if 'http' in self.scheme:
            from tritonclient.http import InferenceServerClient
        elif 'grpc' in self.scheme:
            from tritonclient.grpc import InferenceServerClient
        else:
            raise ValueError(f"Invalid scheme: {self.scheme}")

        url = f'{self.host}:{self.port}'

        return InferenceServerClient(url, concurrency=self.concurrency, connection_timeout=self.connection_timeout,
                                     network_timeout=self.network_timeout, max_greenlets=self.max_greenlets,
                                     ssl_options=self.ssl_options, ssl_context_factory=self.ssl_context_factory,
                                     verbose=self.verbose, insecure=self.insecure, ssl=self.ssl)

    def call_model(self, *args, model_name=None, model_version=None):
        model_name = model_name or self.model_name
        model_version = model_version or self.model_version or ''
        inputs_metadata = []
        org_type = check_type(args[0])
        org_device = args[0].device if org_type.minor == 'tensor' else None
        inputs = []
        metadata = self.get_metadata(model_name, model_version)
        for i, input_metadata in enumerate(metadata['inputs']):
            # input_shape = [1] + input_shape
            input_metadata = self.infer_input(input_metadata['name'], input_metadata['shape'],
                                              input_metadata['datatype'])
            input = args[i]
            input = input_metadata.set_data_from_numpy(
                    as_numpy(input, dtype=triton_to_numpy_dtype_dict[input_metadata.datatype()]))
            inputs_metadata.append(input_metadata)
            inputs.append(input)

        outputs_metadata = []
        for output_metadata in metadata['outputs']:
            output_metadata = self.infer_requested_output(output_metadata['name'])
            outputs_metadata.append(output_metadata)

        response = self.client.infer(model_name=model_name, inputs=inputs, outputs=outputs_metadata)
        outputs = []
        for output_metadata in outputs_metadata:
            out = response.as_numpy(output_metadata.name())
            if org_type.minor == 'tensor':
                out = as_tensor(out, device=org_device)
            outputs.append(out)
            
        if len(outputs) == 1:
            outputs = outputs[0]

        return outputs

    def __call__(self, *args):
        return self.call_model(*args)

    def getattr(self, item):

        if self.model_name is None:
            func = partial(self.call_model, model_name=item)
        else:
            func = partial(self.call_model, model_name=self.model_name, model_version=item)
        return func

    def __getitem__(self, item):
        parts = item.split('/')
        model_name = parts[0]
        model_version = parts[1] if len(parts) > 1 else None
        func = partial(self.call_model, model_name=model_name, model_version=model_version)
        return func



import json
import re

from dataclasses import dataclass, field
from ..path import beam_path, local_copy, BeamURL


pytorch_to_triton_dtype_dict = {
        'float32': 'TYPE_FP32',
        'float16': 'TYPE_FP16',
        'int32': 'TYPE_INT32',
        'int64': 'TYPE_INT64',
        'int16': 'TYPE_INT16',
        'int8': 'TYPE_INT8',
        'uint8': 'TYPE_UINT8',
        'bool': 'TYPE_BOOL',
        'float64': 'TYPE_FP64',
        'bfloat16': 'TYPE_BF16',
    }


def pytorch_to_triton_dtype(dtype):
    dtype = str(dtype)
    return pytorch_to_triton_dtype_dict[dtype]



@dataclass
class TritonConfig:
    name: str = ''
    platform: str = ''
    max_batch_size: int = 0
    input: list = field(default_factory=list)
    output: list = field(default_factory=list)
    instance_groups: list = field(default_factory=list)

    @staticmethod
    def transform_to_json_like(s):
        s = re.sub(r'(\w+)\s*:', r'"\1":', s)  # add quotes for keys
        s = re.sub(r'(\w+)\s+\[', r'"\1": [', s)  # add quotes for keys
        s = re.sub(r'(\w+)\s+{', r'"\1": {', s)  # add quotes for keys
        s = re.sub(r':\s*(?![\["\d])(\w+)', r': "\1"', s)  # add quotes to values
        s = re.sub(r'([}\]"])\s+("\w)', r'\1,\n\2', s)  # Add commas between elements
        s = re.sub(r'(\s+\d+)\s+("\w)', r'\1,\n\2', s)  # Add commas between elements
        s = f"{{{s}}}"
        return s

    @classmethod
    def load_from_file(cls, path):
        path = beam_path(path)
        s = path.read_text()
        s = cls.transform_to_json_like(s)
        parsed_config = json.loads(s)
        return cls(**parsed_config)

    def save_to_file(self, path):
        path = beam_path(path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write(self._serialize_config(), ext='bin')

    def _serialize_config(self):
        config_str = f'name: "{self.name}"\n'
        config_str += f'platform: "{self.platform}"\n'
        for section_name in ['  input', '  output', 'instance_groups']:
            for section in getattr(self, section_name):
                config_str += f'{section_name} [\n'
                config_str += self._serialize_section(section)
                config_str += ']\n'
        return config_str

    @staticmethod
    def _serialize_section(section):
        return '\n'.join([f'  {key}: "{TritonConfig._serialize_value(key, value)}"' for key, value in section.items()])

    @staticmethod
    def _serialize_value(key, value):
        if key == 'dims':
            return '[' + ', '.join(map(str, value)) + ']'
        return value
